# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# There are 2 ways to use this plugin...
# Method 1: Via Webhook:
#   Visit https://my.slack.com/services/new/incoming-webhook/
#   to create a new incoming webhook for your account. You'll need to
#   follow the wizard to pre-determine the channel(s) you want your
#   message to broadcast to, and when you're complete, you will
#   recieve a URL that looks something like this:
#   https://hooks.slack.com/services/T1JJ3T3L2/A1BRTD4JD/TIiajkdnlazkcOXrIdevi7
#                                       ^         ^               ^
#                                       |         |               |
#    These are important <--------------^---------^---------------^
#
# Method 2: Via a Bot:
#   1. visit: https://api.slack.com/apps?new_app=1
#   2. Pick an App Name (such as Apprise) and select your workspace.  Then
#       press 'Create App'
#   3. You'll be able to click on 'Bots' from here where you can then choose
#       to add a 'Bot User'.  Give it a name and choose 'Add Bot User'.
#   4. Now you can choose 'Install App' to which you can choose 'Install App
#       to Workspace'.
#   5. You will need to authorize the app which you get prompted to do.
#   6. Finally you'll get some important information providing you your
#      'OAuth Access Token' and 'Bot User OAuth Access Token' such as:
#        slack://{Oauth Access Token}
#
#        ... which might look something like:
#        slack://xoxp-1234-1234-1234-4ddbc191d40ee098cbaae6f3523ada2d
#        ... or:
#        slack://xoxb-1234-1234-4ddbc191d40ee098cbaae6f3523ada2d
#
#       You must at least give your bot the following access for it to
#       be useful:
#         - chat:write       - MUST be set otherwise you can not post into
#                              a channel
#         - users:read.email - Required if you want to be able to lookup
#                              users by their email address.
#
#      The easiest way to bring a bot into a channel (so that it can send
#      a message to it is to invite it. At this time Apprise does not support
#      an auto-join functionality. To do this:
#        - In the 'Details' section of your channel
#        - Click on the 'More' [...] (elipse icon)
#        - Click 'Add apps'
#        - You will be able to select the Bot App you previously created
#        - Your bot will join your channel.

import contextlib
from json import dumps, loads
from json.decoder import JSONDecodeError
import re
from time import time

import requests

from ..apprise_attachment import AppriseAttachment
from ..common import NotifyFormat, NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import is_email, parse_bool, parse_list, validate_regex
from ..utils.templates import TemplateType, apply_template
from .base import NotifyBase

# Extend HTTP Error Messages
SLACK_HTTP_ERROR_MAP = {
    401: "Unauthorized - Invalid Token.",
}

# Used to break path apart into list of channels
CHANNEL_LIST_DELIM = re.compile(r"[ \t\r\n,#\\/]+")

# Channel Regular Expression Parsing
CHANNEL_RE = re.compile(
    r"^(?P<channel>[+#@]?[A-Z0-9_-]{1,32})(:(?P<thread_ts>[0-9.]+))?$", re.I
)


class SlackMode:
    """Tracks the mode of which we're using Slack."""

    # We're dealing with a webhook
    # Our token looks like: T1JJ3T3L2/A1BRTD4JD/TIiajkdnlazkcOXrIdevi7
    WEBHOOK = "hook"

    # Government Webhook
    # Our token still looks like: T1JJ3T3L2/A1BRTD4JD/TIiajkdnlazkcOXrIdevi7
    # however we have a different URL we post to
    WEBHOOK_GOV = "gov-hook"

    # We're dealing with a bot (using the OAuth Access Token)
    # Our token looks like: xoxp-1234-1234-1234-abc124 or
    # Our token looks like: xoxb-1234-1234-abc124 or
    BOT = "bot"

    # Slack Workflow Builder webhook
    # URL: https://hooks.slack.com/workflows/T.../F.../X.../Y...
    WORKFLOW = "workflow"

    # Slack Workflow trigger webhook (newer format)
    # URL: https://hooks.slack.com/triggers/T.../X.../Y...
    WORKFLOW_TRIGGER = "trigger"


# Define our Slack Modes
SLACK_MODES = (
    SlackMode.WEBHOOK,
    SlackMode.WEBHOOK_GOV,
    SlackMode.BOT,
    SlackMode.WORKFLOW,
    SlackMode.WORKFLOW_TRIGGER,
)


class NotifySlack(NotifyBase):
    """A wrapper for Slack Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Slack"

    # The services URL
    service_url = "https://slack.com/"

    # The default secure protocol
    secure_protocol = "slack"

    # Allow 50 requests per minute (Tier 2).
    # 60/50 = 0.2
    request_rate_per_sec = 1.2

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://appriseit.com/services/slack/"

    # Support attachments
    attachment_support = True

    # The maximum targets to include when doing batch transfers
    # Slack Webhook URLs
    webhook_url = "https://hooks.slack.com/services"
    webhook_gov_url = "https://hooks.slack-gov.com/services"

    # Slack Workflow Builder webhook URL bases
    # See: https://slack.com/help/articles/360041352714
    workflow_url = "https://hooks.slack.com/workflows"
    workflow_trigger_url = "https://hooks.slack.com/triggers"

    # Slack API URL (used with Bots)
    api_url = "https://slack.com/api/{}"

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_72

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 35000

    # There is no reason we should exceed 35KB when reading in a JSON file.
    # If it is more than this, then it is not accepted
    max_slack_template_size = 35000

    # Default Notification Format
    notify_format = NotifyFormat.MARKDOWN

    # Bot's do not have default channels to notify; so #general
    # becomes the default channel in BOT mode
    default_notification_channel = "#general"

    # Define object templates
    templates = (
        # Webhook
        "{schema}://{token_a}/{token_b}/{token_c}",
        "{schema}://{botname}@{token_a}/{token_b}/{token_c}",
        "{schema}://{token_a}/{token_b}/{token_c}/{targets}",
        "{schema}://{botname}@{token_a}/{token_b}/{token_c}/{targets}",
        # Bot
        "{schema}://{access_token}/",
        "{schema}://{access_token}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "botname": {
                "name": _("Bot Name"),
                "type": "string",
                "map_to": "user",
            },
            # Bot User OAuth Access Token
            # which always starts with xoxp- e.g.:
            #     xoxb-1234-1234-4ddbc191d40ee098cbaae6f3523ada2d
            "access_token": {
                "name": _("OAuth Access Token"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^(?:xoxe\.)?xox[abp]-[A-Z0-9-]+$", "i"),
            },
            # Token required as part of the Webhook request
            #  /AAAAAAAAA/........./........................
            "token_a": {
                "name": _("Token A"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[A-Z0-9]+$", "i"),
            },
            # Token required as part of the Webhook request
            #  /........./BBBBBBBBB/........................
            "token_b": {
                "name": _("Token B"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[A-Z0-9]+$", "i"),
            },
            # Token required as part of the Webhook request
            #  /........./........./CCCCCCCCCCCCCCCCCCCCCCCC
            "token_c": {
                "name": _("Token C"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[A-Za-z0-9]+$", "i"),
            },
            "target_encoded_id": {
                "name": _("Target Encoded ID"),
                "type": "string",
                "prefix": "+",
                "map_to": "targets",
            },
            "target_email": {
                "name": _("Target Email"),
                "type": "string",
                "map_to": "targets",
            },
            "target_user": {
                "name": _("Target User"),
                "type": "string",
                "prefix": "@",
                "map_to": "targets",
            },
            "target_channels": {
                "name": _("Target Channel"),
                "type": "string",
                "prefix": "#",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
            # Workflow Builder path: all URL segments joined with slashes
            # e.g. T05XXXX/Ft07XXXX/XXXXXXXX/YYYYYYYY
            "workflow_path": {
                "name": _("Workflow Path"),
                "type": "string",
                "private": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "image": {
                "name": _("Include Image"),
                "type": "bool",
                "default": True,
                "map_to": "include_image",
            },
            "footer": {
                "name": _("Include Footer"),
                "type": "bool",
                "default": True,
                "map_to": "include_footer",
            },
            # Use Payload in Blocks (vs legacy way):
            #  See: https://api.slack.com/reference/messaging/payload
            "blocks": {
                "name": _("Use Blocks"),
                "type": "bool",
                "default": False,
                "map_to": "use_blocks",
            },
            "timestamp": {
                "name": _("Include Timestamp"),
                "type": "bool",
                "default": True,
                "map_to": "include_timestamp",
            },
            "mode": {
                "name": _("Message Mode"),
                "type": "choice:string",
                "values": SLACK_MODES,
                # mode is detected if not specified
            },
            "token": {
                "name": _("Token"),
                "alias_of": ("access_token", "token_a", "token_b", "token_c"),
            },
            "to": {
                "alias_of": "targets",
            },
            "template": {
                "name": _("Template Path"),
                "type": "string",
                "private": True,
            },
        },
    )

    # Define our token control
    template_kwargs = {
        "tokens": {
            "name": _("Template Tokens"),
            "prefix": ":",
        },
    }

    # Formatting requirements are defined here:
    # https://api.slack.com/docs/message-formatting
    _re_formatting_map = {
        # New lines must become the string version
        r"\r\*\n": "\\n",
        # Escape other special characters
        r"&": "&amp;",
        r"<": "&lt;",
        r">": "&gt;",
    }

    # To notify a channel, one uses <!channel|channel>
    _re_channel_support = re.compile(
        r"(?P<match>(?:<|\&lt;)?[ \t]*"
        r"!(?P<channel>[^| \n]+)"
        r"(?:[ \t]*\|[ \t]*(?:(?P<val>[^\n]+?)[ \t]*)?(?:>|\&gt;)"
        r"|(?:>|\&gt;)))",
        re.IGNORECASE,
    )

    # To notify a user by their ID, one uses <@U6TTX1F9R>
    _re_user_id_support = re.compile(
        r"(?P<match>(?:<|\&lt;)?[ \t]*"
        r"@(?P<userid>[^| \n]+)"
        r"(?:[ \t]*\|[ \t]*(?:(?P<val>[^\n]+?)[ \t]*)?(?:>|\&gt;)"
        r"|(?:>|\&gt;)))",
        re.IGNORECASE,
    )

    # The markdown in slack isn't [desc](url), it's <url|desc>
    #
    # To accommodate this, we need to ensure we don't escape URLs that match
    _re_url_support = re.compile(
        r"(?P<match>(?:<|\&lt;)?[ \t]*"
        r"(?P<url>(?:https?|mailto)://[^| \n]+)"
        r"(?:[ \t]*\|[ \t]*(?:(?P<val>[^\n]+?)[ \t]*)?(?:>|\&gt;)"
        r"|(?:>|\&gt;)))",
        re.IGNORECASE,
    )

    def __init__(
        self,
        access_token=None,
        token_a=None,
        token_b=None,
        token_c=None,
        targets=None,
        include_image=None,
        include_footer=None,
        include_timestamp=None,
        use_blocks=None,
        mode=None,
        template=None,
        tokens=None,
        workflow_path=None,
        **kwargs,
    ):
        """Initialize Slack Object."""
        super().__init__(**kwargs)

        # Store our webhook mode
        # Track whether the caller supplied an explicit mode so we can
        # decide later whether to validate exact segment count or
        # auto-detect the workflow sub-mode from segment count.
        _mode_explicit = bool(mode and isinstance(mode, str))
        if _mode_explicit:
            self.mode = next(
                (a for a in SLACK_MODES if a.startswith(mode)), None
            )
            if self.mode not in SLACK_MODES:
                msg = f"The Slack mode specified ({mode}) is invalid."
                self.logger.warning(msg)
                raise TypeError(msg)

        elif workflow_path:
            # Mode will be determined by segment count below;
            # use WORKFLOW as a sentinel so we enter the right branch.
            self.mode = SlackMode.WORKFLOW

        else:  # Detect
            self.mode = SlackMode.BOT if access_token else SlackMode.WEBHOOK

        if self.mode in (
            SlackMode.WORKFLOW,
            SlackMode.WORKFLOW_TRIGGER,
        ):
            # Workflow Builder mode: no OAuth token, no webhook tokens
            self.access_token = None
            self.token_a = None
            self.token_b = None
            self.token_c = None

            # Parse workflow path segments from string or list
            if isinstance(workflow_path, (list, tuple)):
                self.workflow_path = [str(s) for s in workflow_path if s]
            elif isinstance(workflow_path, str):
                self.workflow_path = [s for s in workflow_path.split("/") if s]
            else:
                self.workflow_path = []

            # Segment counts are fixed by Slack:
            #   /workflows/ URLs have exactly 4 segments
            #   /triggers/  URLs have exactly 3 segments
            seg_count = len(self.workflow_path)
            if _mode_explicit:
                # Validate exact count for the declared mode
                expected = 4 if self.mode == SlackMode.WORKFLOW else 3
                if seg_count != expected:
                    msg = (
                        f"A Slack {self.mode!r} URL requires exactly"
                        f" {expected} path segments"
                        f" ({workflow_path!r} has {seg_count})."
                    )
                    self.logger.warning(msg)
                    raise TypeError(msg)
            else:
                # Auto-detect sub-mode from segment count
                if seg_count == 4:
                    self.mode = SlackMode.WORKFLOW
                elif seg_count == 3:
                    self.mode = SlackMode.WORKFLOW_TRIGGER
                else:
                    msg = (
                        "A Slack Workflow URL requires exactly 3 or 4"
                        f" path segments ({workflow_path!r} has"
                        f" {seg_count})."
                    )
                    self.logger.warning(msg)
                    raise TypeError(msg)

        elif self.mode in (SlackMode.WEBHOOK, SlackMode.WEBHOOK_GOV):
            self.workflow_path = []
            self.access_token = None
            self.token_a = validate_regex(
                token_a, *self.template_tokens["token_a"]["regex"]
            )
            if not self.token_a:
                msg = (
                    "An invalid Slack (first) Token "
                    f"({token_a}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg)

            self.token_b = validate_regex(
                token_b, *self.template_tokens["token_b"]["regex"]
            )
            if not self.token_b:
                msg = (
                    "An invalid Slack (second) Token "
                    f"({token_b}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg)

            self.token_c = validate_regex(
                token_c, *self.template_tokens["token_c"]["regex"]
            )
            if not self.token_c:
                msg = (
                    "An invalid Slack (third) Token "
                    f"({token_c}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg)
        else:
            self.workflow_path = []
            self.token_a = None
            self.token_b = None
            self.token_c = None
            self.access_token = validate_regex(
                access_token, *self.template_tokens["access_token"]["regex"]
            )
            if not self.access_token:
                msg = (
                    "An invalid Slack OAuth Access Token "
                    f"({access_token}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg)

        # Look the users up by their email address and map them back to their
        # id here for future queries (if needed). This allows people to
        # specify a full email as a recipient via slack
        self._lookup_users = {}

        self.use_blocks = (
            parse_bool(use_blocks, self.template_args["blocks"]["default"])
            if use_blocks is not None
            else self.template_args["blocks"]["default"]
        )

        # Build list of channels
        self.channels = parse_list(targets)
        if len(self.channels) == 0:
            # No problem; the webhook is smart enough to just notify the
            # channel it was created for; adding 'None' is just used as
            # a flag lower to not set the channels
            self.channels.append(
                None
                if self.mode
                in (
                    SlackMode.WEBHOOK,
                    SlackMode.WEBHOOK_GOV,
                    # Workflow mode has no channel concept
                    SlackMode.WORKFLOW,
                    SlackMode.WORKFLOW_TRIGGER,
                )
                else self.default_notification_channel
            )

        # Iterate over above list and store content accordingly
        self._re_formatting_rules = re.compile(
            r"(" + "|".join(self._re_formatting_map.keys()) + r")",
            re.IGNORECASE,
        )
        # Place a thumbnail image inline with the message body
        self.include_image = (
            self.template_args["image"]["default"]
            if include_image is None
            else include_image
        )

        # Place a footer with each post
        self.include_footer = (
            self.template_args["footer"]["default"]
            if include_footer is None
            else include_footer
        )

        # timestamp inclusion (only applicable if footer also defined
        self.include_timestamp = (
            self.template_args["timestamp"]["default"]
            if include_timestamp is None
            else include_timestamp
        )

        # Our template object is just an AppriseAttachment object
        self.template = AppriseAttachment(asset=self.asset)
        if template:
            # Add our definition to our template
            self.template.add(template)
            if not len(self.template):
                # add() failed (unsupported schema, unparseable URL, etc.)
                msg = f"The Slack template ({template!r}) could not be loaded."
                self.logger.warning(msg)
                raise TypeError(msg)
            # Enforce maximum file size
            self.template[0].max_file_size = self.max_slack_template_size

        # Template functionality
        self.tokens = {}
        if isinstance(tokens, dict):
            self.tokens.update(tokens)

        elif tokens:
            msg = (
                "The specified Slack Template Tokens "
                f"({tokens}) are not identified as a dictionary."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # A template always implies Block Kit mode; there is no meaningful
        # use for templates outside of blocks mode.
        if self.template:
            self.use_blocks = True

        return

    def gen_payload(
        self, body, title="", notify_type=NotifyType.INFO, **kwargs
    ):
        """Generate our payload from an external Block Kit JSON template.

        Returns the validated attachment content dict (which must contain
        a non-empty 'blocks' list with typed entries) to be placed into
        the Slack attachments array.  Returns False on any failure.
        """
        # Acquire the first template attachment
        template = self.template[0]
        if not template:
            # We could not access the attachment
            self.logger.error(
                "Could not access Slack template"
                f" {template.url(privacy=True)}."
            )
            return False

        # Acquire our to-be footer icon if configured to do so
        image_url = (
            None if not self.include_image else self.image_url(notify_type)
        )

        # Take a copy of our token dictionary
        tokens = self.tokens.copy()

        # Apply some defaults template values
        tokens["app_body"] = body
        tokens["app_title"] = title
        tokens["app_type"] = notify_type.value
        tokens["app_id"] = self.app_id
        tokens["app_desc"] = self.app_desc
        tokens["app_color"] = self.color(notify_type)
        tokens["app_image_url"] = image_url
        tokens["app_url"] = self.app_url

        # Templates are always Block Kit JSON; enforce JSON escaping
        tokens["app_mode"] = TemplateType.JSON

        # Coerce all substitution values to str before JSON-escaping.
        # apply_template's _escape_json() calls json.dumps(v)[1:-1] which
        # is only correct for strings; None produces "ul" and other non-
        # string types produce similarly corrupted output.  app_mode is a
        # TemplateType sentinel passed as a named parameter, not a
        # substitution value, so it is left as-is.
        safe_tokens = {
            k: (
                v
                if k == "app_mode" or isinstance(v, str)
                else ("" if v is None else str(v))
            )
            for k, v in tokens.items()
        }

        try:
            with open(template.path) as fp:
                content = apply_template(fp.read(), **safe_tokens)

        except OSError:
            self.logger.error(
                "Slack template"
                f" {template.url(privacy=True)} could not be read."
            )
            return False

        # Parse and validate as JSON
        try:
            content = loads(content)

        except JSONDecodeError as e:
            self.logger.error(
                "Slack template"
                f" {template.url(privacy=True)} contains invalid JSON."
            )
            self.logger.debug(f"JSONDecodeError: {e}")
            return False

        # Template must parse to a JSON object, not an array or scalar
        if not isinstance(content, dict):
            self.logger.error(
                "Slack template"
                f" {template.url(privacy=True)} must be a JSON object"
                " (got {}).".format(type(content).__name__)
            )
            return False

        # 'blocks' must be a non-empty list (Block Kit requirement)
        if (
            not isinstance(content.get("blocks"), list)
            or not content["blocks"]
        ):
            self.logger.error(
                "Slack template"
                f" {template.url(privacy=True)} must contain"
                " a non-empty 'blocks' list."
            )
            return False

        # Every block must be a dict with a 'type' string (Block Kit)
        if not all(
            isinstance(b, dict) and isinstance(b.get("type"), str)
            for b in content["blocks"]
        ):
            self.logger.error(
                "Slack template"
                f" {template.url(privacy=True)} contains"
                " a block missing a 'type' string."
            )
            return False

        # Return the validated attachment content dict
        return content

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform Slack Notification."""

        # error tracking (used for function return)
        has_error = False

        #
        # Workflow Builder mode: fixed endpoint, no channel iteration
        #
        if self.mode in (SlackMode.WORKFLOW, SlackMode.WORKFLOW_TRIGGER):
            if self.template:
                # Use external Block Kit JSON template
                payload = self.gen_payload(
                    body, title, notify_type=notify_type, **kwargs
                )
                if payload is False:
                    return False

            else:
                # Default payload: combine title and body into a single
                # 'text' field -- the workflow must accept this variable.
                payload = {
                    "text": f"{title}: {body}" if title else body,
                }

            # Construct the workflow POST URL from stored path segments
            wf_base = (
                self.workflow_trigger_url
                if self.mode is SlackMode.WORKFLOW_TRIGGER
                else self.workflow_url
            )
            url = "{}/{}".format(wf_base, "/".join(self.workflow_path))
            # _send() returns False on error, non-False on success
            return self._send(url, payload) is not False

        #
        # Prepare JSON Object (applicable to both WEBHOOK and BOT mode)
        #
        if self.use_blocks:
            if self.template:
                # Template mode: generate blocks attachment from template
                template_content = self.gen_payload(
                    body, title, notify_type=notify_type, **kwargs
                )
                if template_content is False:
                    # Template loading or parsing failed; abort early
                    return False

                # Wrap the template attachment content in a full payload
                payload = {
                    "username": self.user if self.user else self.app_id,
                    "attachments": [template_content],
                }

            else:
                # Our slack format
                slack_format = (
                    "mrkdwn"
                    if self.notify_format == NotifyFormat.MARKDOWN
                    else "plain_text"
                )

                payload = {
                    "username": self.user if self.user else self.app_id,
                    "attachments": [
                        {
                            "blocks": [
                                {
                                    "type": "section",
                                    "text": {
                                        "type": slack_format,
                                        "text": body,
                                    },
                                }
                            ],
                            "color": self.color(notify_type),
                        }
                    ],
                }

                # Slack only accepts non-empty header sections
                if title:
                    payload["attachments"][0]["blocks"].insert(
                        0,
                        {
                            "type": "header",
                            "text": {
                                "type": "plain_text",
                                "text": title,
                                "emoji": True,
                            },
                        },
                    )

                # Include the footer only if specified to do so
                if self.include_footer:
                    # Acquire our to-be footer icon if configured to do so
                    image_url = (
                        None
                        if not self.include_image
                        else self.image_url(notify_type)
                    )

                    # Prepare our footer based on the block structure
                    footer = {
                        "type": "context",
                        "elements": [
                            {"type": slack_format, "text": self.app_id}
                        ],
                    }

                    if image_url:
                        payload["icon_url"] = image_url

                        footer["elements"].insert(
                            0,
                            {
                                "type": "image",
                                "image_url": image_url,
                                "alt_text": notify_type,
                            },
                        )

                    payload["attachments"][0]["blocks"].append(footer)

        else:
            #
            # Legacy API Formatting
            #
            if self.notify_format == NotifyFormat.MARKDOWN:
                body = self._re_formatting_rules.sub(  # pragma: no branch
                    lambda x: self._re_formatting_map[x.group()],
                    body,
                )

                # Support <!channel|desc>, <!channel> entries
                for match in self._re_channel_support.findall(body):
                    # Swap back any ampersands previously updaated
                    channel = match[1].strip()
                    desc = match[2].strip()

                    # Update our string
                    body = re.sub(
                        re.escape(match[0]),
                        f"<!{channel}|{desc}>" if desc else f"<!{channel}>",
                        body,
                        flags=re.IGNORECASE,
                    )

                # Support <@userid|desc>, <@channel> entries
                for match in self._re_user_id_support.findall(body):
                    # Swap back any ampersands previously updaated
                    user = match[1].strip()
                    desc = match[2].strip()

                    # Update our string
                    body = re.sub(
                        re.escape(match[0]),
                        f"<@{user}|{desc}>" if desc else f"<@{user}>",
                        body,
                        flags=re.IGNORECASE,
                    )

                # Support <url|desc>, <url> entries
                for match in self._re_url_support.findall(body):
                    # Swap back any ampersands previously updaated
                    url = match[1].replace("&amp;", "&")
                    desc = match[2].strip()

                    # Update our string
                    body = re.sub(
                        re.escape(match[0]),
                        f"<{url}|{desc}>" if desc else f"<{url}>",
                        body,
                        flags=re.IGNORECASE,
                    )

            # Perform Formatting on title here; this is not needed for block
            # mode above
            title = self._re_formatting_rules.sub(  # pragma: no branch
                lambda x: self._re_formatting_map[x.group()],
                title,
            )

            # Prepare JSON Object (applicable to both WEBHOOK and BOT mode)
            payload = {
                "username": self.user if self.user else self.app_id,
                # Use Markdown language
                "mrkdwn": self.notify_format == NotifyFormat.MARKDOWN,
                "attachments": [
                    {
                        "title": title,
                        "text": body,
                        "color": self.color(notify_type),
                    }
                ],
            }

            # Acquire our to-be footer icon if configured to do so
            image_url = (
                None if not self.include_image else self.image_url(notify_type)
            )

            if image_url:
                payload["icon_url"] = image_url

            # Include the footer only if specified to do so
            if self.include_footer:
                if image_url:
                    payload["attachments"][0]["footer_icon"] = image_url

                # Include the footer only if specified to do so
                payload["attachments"][0]["footer"] = self.app_id

                if self.include_timestamp:
                    # Timestamp
                    payload["attachments"][0]["ts"] = time()
        if (
            attach
            and self.attachment_support
            and self.mode in (SlackMode.WEBHOOK, SlackMode.WEBHOOK_GOV)
        ):
            # Be friendly; let the user know why they can't send their
            # attachments if using the Webhook mode
            self.logger.warning("Slack Webhooks do not support attachments.")

        # Prepare our Slack URL (depends on mode)
        if self.mode is SlackMode.WEBHOOK:
            url = (
                f"{self.webhook_url}/{self.token_a}"
                f"/{self.token_b}/{self.token_c}"
            )

        elif self.mode is SlackMode.WEBHOOK_GOV:
            url = (
                f"{self.webhook_gov_url}/{self.token_a}"
                f"/{self.token_b}/{self.token_c}"
            )

        else:  # SlackMode.BOT
            url = self.api_url.format("chat.postMessage")

        # Create a copy of the channel list
        channels = list(self.channels)

        attach_channel_list = []
        while len(channels):
            channel = channels.pop(0)
            if channel is not None:
                # We'll perform a user lookup if we detect an email
                email = is_email(channel)
                if email:
                    payload["channel"] = self.lookup_userid(
                        email["full_email"]
                    )

                    if not payload["channel"]:
                        # Move along; any notifications/logging would have
                        # come from lookup_userid()
                        has_error = True
                        continue

                else:  # Channel
                    result = CHANNEL_RE.match(channel)

                    if not result:
                        # Channel over-ride was specified
                        self.logger.warning(
                            f"The specified Slack target {channel} is invalid;"
                            "skipping."
                        )

                        # Mark our failure
                        has_error = True
                        continue

                    # Store oure content
                    channel, thread_ts = (
                        result.group("channel"),
                        result.group("thread_ts"),
                    )
                    if thread_ts:
                        payload["thread_ts"] = thread_ts

                    elif "thread_ts" in payload:
                        # Handle situations where one channel has a thread_id
                        # specified, and the next does not.  We do not want to
                        # cary forward the last value specified
                        del payload["thread_ts"]

                    if channel[0] == "+":
                        # Treat as encoded id if prefixed with a +
                        payload["channel"] = channel[1:]

                    elif channel[0] == "@":
                        # Treat @ value 'as is'
                        payload["channel"] = channel

                    else:
                        # Prefix with channel hash tag (if not already)
                        payload["channel"] = (
                            channel if channel[0] == "#" else f"#{channel}"
                        )

            response = self._send(url, payload)
            if not response:
                # Handle any error
                has_error = True
                continue

            # Store the valid channel or chat ID (for DMs) that will
            # be accepted by Slack's attachment method later.
            if response.get("channel"):
                attach_channel_list.append(response.get("channel"))

            self.logger.info(
                "Sent Slack notification{}.".format(
                    f" to {channel}" if channel is not None else ""
                )
            )

        if (
            attach
            and self.attachment_support
            and self.mode is SlackMode.BOT
            and attach_channel_list
        ):
            # Send our attachments (can only be done in bot mode)
            for no, attachment in enumerate(attach, start=1):
                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                self.logger.debug(
                    f"Posting Slack attachment {attachment.url(privacy=True)}"
                )

                # Get the URL to which to upload the file.
                # https://api.slack.com/methods/files.getUploadURLExternal
                params = {
                    "filename": (
                        attachment.name
                        if attachment.name
                        else f"file{no:03}.dat"
                    ),
                    "length": len(attachment),
                }
                url_ = self.api_url.format("files.getUploadURLExternal")
                response = self._send(
                    url_, {}, http_method="get", params=params
                )
                if not (
                    response
                    and response.get("file_id")
                    and response.get("upload_url")
                ):
                    self.logger.error("Could retrieve file upload URL.")
                    # We failed to get an upload URL, take an early exit
                    return False

                file_id = response.get("file_id")
                upload_url = response.get("upload_url")

                # Upload file
                response = self._send(upload_url, {}, attach=attachment)

                # Send file to channels
                # https://api.slack.com/methods/files.completeUploadExternal
                for channel_id in attach_channel_list:
                    payload_ = {
                        "files": [
                            {
                                "id": file_id,
                                "title": attachment.name,
                            }
                        ],
                        "channel_id": channel_id,
                    }
                    url_ = self.api_url.format("files.completeUploadExternal")
                    response = self._send(url_, payload_)
                    # Expected response
                    # {
                    #     "ok": true,
                    #     "files": [
                    #         {
                    #             "id": "F123ABC456",
                    #             "title": "slack-test"
                    #         }
                    #     ]
                    # }
                    if not (response and response.get("files")):
                        self.logger.error("Failed to send file to channel.")
                        # We failed to send the file to the channel,
                        # take an early exit
                        return False

        return not has_error

    def lookup_userid(self, email):
        """Takes an email address and attempts to resolve/acquire it's user id
        for notification purposes."""
        if email in self._lookup_users:
            # We're done as entry has already been retrieved
            return self._lookup_users[email]

        if self.mode is not SlackMode.BOT:
            # You can not look up
            self.logger.warning(
                "Emails can not be resolved to Slack User IDs unless you "
                "have a bot configured."
            )
            return None

        lookup_url = self.api_url.format("users.lookupByEmail")
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": f"Bearer {self.access_token}",
        }

        # we pass in our email address as the argument
        params = {
            "email": email,
        }

        self.logger.debug(
            "Slack User Lookup POST URL:"
            f" {lookup_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Slack User Lookup Parameters: {params!s}")

        # Initialize our HTTP JSON response
        response = {"ok": False}

        # Initialize our detected user id (also the response to this function)
        user_id = None

        # Always call throttle before any remote server i/o is made
        self.throttle()
        try:
            r = requests.get(
                lookup_url,
                headers=headers,
                params=params,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
                allow_redirects=self.redirects,
            )

            # Attachment posts return a JSON string
            with contextlib.suppress(AttributeError, TypeError, ValueError):
                # Load our JSON object if we can

                # ValueError = r.content is Unparsable
                # TypeError = r.content is None
                # AttributeError = r is None
                response = loads(r.content)

            # We can get a 200 response, but still fail.  A failure message
            # might look like this (missing bot permissions):
            #    {
            #      'ok': False,
            #      'error': 'missing_scope',
            #      'needed': 'users:read.email',
            #      'provided': 'calls:write,chat:write'
            #    }

            if r.status_code != requests.codes.ok or not (
                response and response.get("ok", False)
            ):
                # We had a problem
                status_str = NotifySlack.http_response_code_lookup(
                    r.status_code, SLACK_HTTP_ERROR_MAP
                )

                self.logger.warning(
                    "Failed to send Slack User Lookup:{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(
                    "Response Details:\r\n%r", (r.content or b"")[:2000]
                )

                # Return; we're done
                return False

            # If we reach here, then we were successful in looking up
            # the user. A response generally looks like this:
            # {
            #   'ok': True,
            #   'user': {
            #     'id': 'J1ZQB9T9Y',
            #     'team_id': 'K1WR6TML2',
            #     'name': 'l2g',
            #     'deleted': False,
            #     'color': '9f69e7',
            #     'real_name': 'Chris C',
            #     'tz': 'America/New_York',
            #     'tz_label': 'Eastern Standard Time',
            #     'tz_offset': -18000,
            #     'profile': {
            #       'title': '',
            #       'phone': '',
            #       'skype': '',
            #       'real_name': 'Chris C',
            #       'real_name_normalized':
            #       'Chris C',
            #       'display_name': 'l2g',
            #       'display_name_normalized': 'l2g',
            #       'fields': None,
            #       'status_text': '',
            #       'status_emoji': '',
            #       'status_expiration': 0,
            #       'avatar_hash': 'g785e9c0ddf6',
            #       'email': 'lead2gold@gmail.com',
            #       'first_name': 'Chris',
            #       'last_name': 'C',
            #       'image_24': 'https://secure.gravatar.com/...',
            #       'image_32': 'https://secure.gravatar.com/...',
            #       'image_48': 'https://secure.gravatar.com/...',
            #       'image_72': 'https://secure.gravatar.com/...',
            #       'image_192': 'https://secure.gravatar.com/...',
            #       'image_512': 'https://secure.gravatar.com/...',
            #       'status_text_canonical': '',
            #       'team': 'K1WR6TML2'
            #     },
            #     'is_admin': True,
            #     'is_owner': True,
            #     'is_primary_owner': True,
            #     'is_restricted': False,
            #     'is_ultra_restricted': False,
            #     'is_bot': False,
            #     'is_app_user': False,
            #     'updated': 1603904274
            #   }
            # }
            # We're only interested in the id
            user_id = response["user"]["id"]

            # Cache it for future
            self._lookup_users[email] = user_id
            self.logger.info(
                "Email %s resolves to the Slack User ID: %s.", email, user_id
            )

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred looking up Slack User.",
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            # Return; we're done
            return None

        return user_id

    def _send(
        self,
        url,
        payload,
        attach=None,
        http_method="post",
        params=None,
        **kwargs,
    ):
        """Wrapper to the requests (post) object."""
        self.logger.debug(
            f"Slack POST URL: {url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Slack Payload: {payload!s}")

        headers = {
            "User-Agent": self.app_id,
            "Accept": "application/json",
        }

        if not attach:
            headers["Content-Type"] = "application/json; charset=utf-8"

        if self.mode is SlackMode.BOT:
            headers["Authorization"] = f"Bearer {self.access_token}"

        # Our response object
        response = {"ok": False}

        # Always call throttle before any remote server i/o is made
        self.throttle()

        # Our attachment path (if specified)
        files = None

        try:
            # Open our attachment path if required:
            if attach:
                files = {
                    "file": (
                        attach.name,
                        # file handle is safely closed in `finally`; inline
                        # open is intentional; attach.open() dispatches to
                        # BytesIO for memory attachments
                        attach.open(),
                    ),
                }

            r = requests.request(
                http_method,
                url,
                data=payload if attach else dumps(payload),
                headers=headers,
                files=files,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
                allow_redirects=self.redirects,
                params=params if params else None,
            )

            # Posts return a JSON string
            with contextlib.suppress(AttributeError, TypeError, ValueError):
                # Load our JSON object if we can
                # ValueError = r.content is Unparsable
                # TypeError = r.content is None
                # AttributeError = r is None
                response = loads(r.content)

            # Another response type is:
            # {
            #   'ok': False,
            #   'error': 'not_in_channel',
            # }
            status_okay = False
            if self.mode is SlackMode.BOT:
                status_okay = (
                    (response and response.get("ok", False))
                    or
                    # Responses for file uploads look like this
                    # 'OK - <file length>'
                    (
                        r.content
                        and isinstance(r.content, bytes)
                        and b"OK" in r.content
                    )
                )
            elif self.mode in (
                SlackMode.WORKFLOW,
                SlackMode.WORKFLOW_TRIGGER,
            ):
                # Workflow trigger webhooks confirm success via HTTP 200;
                # the response body may be empty or non-standard JSON.
                status_okay = r.status_code == requests.codes.ok
            elif r.content == b"ok":
                # The text 'ok' is returned if this is a Webhook request
                # So the below captures that as well.
                status_okay = True

            if r.status_code != requests.codes.ok or not status_okay:
                # We had a problem
                status_str = NotifySlack.http_response_code_lookup(
                    r.status_code, SLACK_HTTP_ERROR_MAP
                )

                self.logger.warning(
                    "Failed to send{} to Slack: {}{}error={}.".format(
                        (" " + attach.name) if attach else "",
                        status_str,
                        ", " if status_str else "",
                        r.status_code,
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")
                return False

            # Message Post Response looks like this:
            # {
            #   "attachments": [
            #     {
            #       "color": "3AA3E3",
            #       "fallback": "test",
            #       "id": 1,
            #       "text": "my body",
            #       "title": "my title",
            #       "ts": 1573694687
            #     }
            #   ],
            #   "bot_id": "BAK4K23G5",
            #   "icons": {
            #     "image_48": "https://s3-us-west-2.amazonaws.com/...
            #   },
            #   "subtype": "bot_message",
            #   "text": "",
            #   "ts": "1573694689.003700",
            #   "type": "message",
            #   "username": "Apprise"
            # }

            # files.completeUploadExternal responses look like this:
            # {
            #     "ok": true,
            #     "files": [
            #         {
            #             "id": "F123ABC456",
            #             "title": "slack-test"
            #         }
            #     ]
            # }
        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred posting {}to Slack.".format(
                    attach.name if attach else ""
                )
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            return False

        except OSError as e:
            self.logger.warning(
                "An I/O error occurred while reading {}.".format(
                    attach.name if attach else "attachment"
                )
            )
            self.logger.debug(f"I/O Exception: {e!s}")
            return False

        finally:
            # Close our file (if it's open) stored in the second element
            # of our files tuple (index 1)
            if files:
                files["file"][1].close()

        # Return the response for processing
        return response

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        if self.mode in (SlackMode.WORKFLOW, SlackMode.WORKFLOW_TRIGGER):
            # Workflow mode is identified by its path segments
            return (self.secure_protocol, self.mode, *self.workflow_path)

        return (
            self.secure_protocol,
            self.token_a,
            self.token_b,
            self.token_c,
            self.access_token,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "image": "yes" if self.include_image else "no",
            "footer": "yes" if self.include_footer else "no",
            "timestamp": "yes" if self.include_timestamp else "no",
            "blocks": "yes" if self.use_blocks else "no",
            "mode": self.mode,
        }

        if self.template:
            params["template"] = NotifySlack.quote(
                self.template[0].url(), safe=""
            )

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Store any template token entries if specified
        params.update({f":{k}": v for k, v in self.tokens.items()})

        # Determine if there is a botname present
        botname = ""
        if self.user:
            botname = "{botname}@".format(
                botname=NotifySlack.quote(self.user, safe=""),
            )

        if self.mode in (SlackMode.WORKFLOW, SlackMode.WORKFLOW_TRIGGER):
            return "{schema}://{path}/?{params}".format(
                schema=self.secure_protocol,
                path="/".join(
                    self.pprint(s, privacy, safe="")
                    for s in self.workflow_path
                ),
                params=NotifySlack.urlencode(params),
            )

        if self.mode in (SlackMode.WEBHOOK, SlackMode.WEBHOOK_GOV):
            return (
                "{schema}://{botname}{token_a}/{token_b}/{token_c}/"
                "{targets}/?{params}".format(
                    schema=self.secure_protocol,
                    botname=botname,
                    token_a=self.pprint(self.token_a, privacy, safe=""),
                    token_b=self.pprint(self.token_b, privacy, safe=""),
                    token_c=self.pprint(self.token_c, privacy, safe=""),
                    targets="/".join(
                        [NotifySlack.quote(x, safe="") for x in self.channels]
                    ),
                    params=NotifySlack.urlencode(params),
                )
            )
        # else -> self.mode == SlackMode.BOT:
        return "{schema}://{botname}{access_token}/{targets}/?{params}".format(
            schema=self.secure_protocol,
            botname=botname,
            access_token=self.pprint(self.access_token, privacy, safe=""),
            targets="/".join(
                [NotifySlack.quote(x, safe="") for x in self.channels]
            ),
            params=NotifySlack.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.channels)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # The first token is stored in the hostname
        token = NotifySlack.unquote(results["host"])

        # Get unquoted path entries
        entries = NotifySlack.split_path(results["fullpath"])

        # Peek at mode early to guide path parsing.
        # Resolve via the same prefix-matching used in __init__ so that
        # abbreviated inputs (e.g. mode=tri or mode=work) are handled
        # identically in both places.
        _mode_raw = results["qsd"].get("mode", "")
        _mode = (
            next(
                (a for a in SLACK_MODES if a.startswith(_mode_raw)),
                "",
            )
            if _mode_raw
            else ""
        )

        if _mode in (SlackMode.WORKFLOW, SlackMode.WORKFLOW_TRIGGER):
            # All path segments form the workflow_path
            results["workflow_path"] = "/".join([token, *entries])
            results["targets"] = []

        # Verify if our token is a bot token or part of a webhook:
        elif token.startswith("xo"):
            # We're dealing with a bot
            results["access_token"] = token
            results["targets"] = entries

        else:
            # We're dealing with a webhook
            results["token_a"] = token
            results["token_b"] = entries.pop(0) if entries else None
            results["token_c"] = entries.pop(0) if entries else None
            # assign remaining entries to the channels we wish to notify
            results["targets"] = entries

        # Support the token flag where you can set it to the bot token
        # or the webhook token (with slash delimiters)
        if "token" in results["qsd"] and len(results["qsd"]["token"]):
            # Break our entries up into a list; we can ue the Channel
            # list delimiter above since it doesn't contain any characters
            # we don't otherwise accept anyway in our token
            entries = list(
                filter(
                    bool,
                    CHANNEL_LIST_DELIM.split(
                        NotifySlack.unquote(results["qsd"]["token"])
                    ),
                )
            )

            # check to see if we're dealing with a bot/user token
            if entries and entries[0].startswith("xo"):
                # We're dealing with a bot
                results["access_token"] = entries[0]
                results["token_a"] = None
                results["token_b"] = None
                results["token_c"] = None

            else:  # Webhook
                results["access_token"] = None
                results["token_a"] = entries.pop(0) if entries else None
                results["token_b"] = entries.pop(0) if entries else None
                results["token_c"] = entries.pop(0) if entries else None

        # Support the 'to' variable so that we can support rooms this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += list(
                filter(
                    bool,
                    CHANNEL_LIST_DELIM.split(
                        NotifySlack.unquote(results["qsd"]["to"])
                    ),
                )
            )

        # Get Image Flag
        results["include_image"] = parse_bool(
            results["qsd"].get(
                "image", NotifySlack.template_args["image"]["default"]
            )
        )

        results["include_timestamp"] = parse_bool(
            results["qsd"].get(
                "timestamp", NotifySlack.template_args["timestamp"]["default"]
            )
        )

        # Get Payload structure (use blocks?)
        if "blocks" in results["qsd"] and len(results["qsd"]["blocks"]):
            results["use_blocks"] = parse_bool(results["qsd"]["blocks"])

        # Get Footer Flag
        results["include_footer"] = parse_bool(
            results["qsd"].get(
                "footer", NotifySlack.template_args["footer"]["default"]
            )
        )

        # Get Mode
        if "mode" in results["qsd"] and len(results["qsd"]["mode"]):
            results["mode"] = NotifySlack.unquote(results["qsd"]["mode"])

        # Template Handling
        if "template" in results["qsd"] and results["qsd"]["template"]:
            results["template"] = NotifySlack.unquote(
                results["qsd"]["template"]
            )

        # Store our tokens
        results["tokens"] = results["qsd:"]

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Supports:
          - https://hooks.slack.com/services/TOKEN_A/TOKEN_B/TOKEN_C
          - https://hooks.slack-gov.com/services/TOKEN_A/TOKEN_B/TOKEN_C
          - https://hooks.slack.com/workflows/T.../F.../X.../Y...
          - https://hooks.slack.com/triggers/T.../X.../Y...
        """

        # Workflow Builder and trigger webhook URLs
        wf_result = re.match(
            r"^https?://hooks\.slack\.com/"
            r"(?P<wf_type>workflows|triggers)/"
            r"(?P<path>[A-Z0-9][A-Z0-9/_-]+)/?"
            r"(?P<params>\?.+)?$",
            url,
            re.I,
        )
        if wf_result:
            wf_type = wf_result.group("wf_type").lower()
            # Map URL path type to Apprise mode string
            mode = (
                SlackMode.WORKFLOW_TRIGGER
                if wf_type == "triggers"
                else SlackMode.WORKFLOW
            )
            path = wf_result.group("path").strip("/")
            params = wf_result.group("params") or ""
            sep = "&" if params else "?"
            return NotifySlack.parse_url(
                "{schema}://{path}/{params}{sep}mode={mode}".format(
                    schema=NotifySlack.secure_protocol,
                    path=path,
                    params=params,
                    sep=sep,
                    mode=mode,
                )
            )

        # Standard incoming webhook and government webhook URLs
        result = re.match(
            r"^https?://(?P<host>hooks\.slack(?P<gov>-gov)?\.com)/services/"
            r"(?P<token_a>[A-Z0-9]+)/"
            r"(?P<token_b>[A-Z0-9]+)/"
            r"(?P<token_c>[A-Z0-9]+)/?"
            r"(?P<params>\?.+)?$",
            url,
            re.I,
        )

        if result:
            params = (
                "" if not result.group("params") else result.group("params")
            )

            if result.group("gov"):
                # provide gov parameters
                params = (
                    "?" if not params else "&"
                ) + f"mode={SlackMode.WEBHOOK_GOV}"

            return NotifySlack.parse_url(
                "{schema}://{token_a}/{token_b}/{token_c}/{params}".format(
                    schema=NotifySlack.secure_protocol,
                    token_a=result.group("token_a"),
                    token_b=result.group("token_b"),
                    token_c=result.group("token_c"),
                    params=params,
                )
            )

        return None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from datetime import datetime, timedelta, timezone
import hashlib
import os

from ..apprise_attachment import AppriseAttachment
from ..asset import AppriseAsset
from ..exception import ApprisePluginException
from ..logger import logger


def _ensure_imghdr_shim():
    """Install a minimal imghdr shim when the module is absent.

    pgpy 0.6.0 imports imghdr, which was removed from the standard
    library in Python 3.13 (PEP 594). pgpy's only use of imghdr is
    ImageEncoding.encodingof(), which Apprise never calls. Returning
    None from what() is the same safe fallback the library uses
    internally for non-JPEG data.
    """
    try:
        import imghdr  # noqa: F401
    except ImportError:
        import sys
        import types

        # Build a minimal stand-in with the single function pgpy calls
        shim = types.ModuleType("imghdr")
        shim.what = lambda file=None, h=None: None
        sys.modules["imghdr"] = shim


# Install shim before pgpy attempts its top-level import of imghdr
_ensure_imghdr_shim()

try:
    import pgpy

    # Pretty Good Privacy (PGP) Support enabled
    PGP_SUPPORT = True

except ImportError:
    # Pretty Good Privacy (PGP) Support disabled
    PGP_SUPPORT = False


class ApprisePGPException(ApprisePluginException):
    """Thrown when there is an error with the Pretty Good Privacy
    Controller."""

    def __init__(self, message, error_code=602):
        super().__init__(message, error_code=error_code)


class ApprisePGPController:
    """Pretty Good Privacy Controller Tool for the Apprise Library."""

    # There is no reason a PGP Public Key should exceed 8K in size
    # If it is more than this, then it is not accepted
    max_pgp_public_key_size = 8000

    # Private keys can be materially larger than public keys, especially for
    # 4096-bit RSA or keys carrying multiple subkeys and UIDs; 32K is
    # generous enough to accommodate all realistic cases without being
    # unbounded.
    max_pgp_private_key_size = 32000

    def __init__(
        self,
        path,
        pub_keyfile=None,
        prv_keyfile=None,
        email=None,
        asset=None,
        wkd=None,
        **kwargs,
    ):
        """Path should be the directory keys can be written and read from such
        as <notifyobject>.store.path.

        Optionally additionally specify a pub_keyfile and/or prv_keyfile to
        use explicit key files, and/or an AppriseWKDController to enable Web
        Key Directory key discovery when no local key is found.
        """

        # PGP hash
        self.__key_lookup = {}

        # Directory we can work with
        self.path = path

        # Our email
        self.email = email

        # Optional WKD controller for automatic key discovery
        self.wkd = wkd

        # Prepare our Asset Object
        self.asset = (
            asset if isinstance(asset, AppriseAsset) else AppriseAsset()
        )

        if pub_keyfile:
            # Create ourselves an Attachment to work with; this grants us the
            # ability to pull this key from a remote site or anything else
            # supported by the Attachment object
            self._pub_keyfile = AppriseAttachment(asset=self.asset)

            # Add our definition to our pgp_key reference
            self._pub_keyfile.add(pub_keyfile)

            # Enforce maximum file size
            self._pub_keyfile[0].max_file_size = self.max_pgp_public_key_size

        else:
            self._pub_keyfile = None

        if prv_keyfile:
            # Create an Attachment for the private key as well so that remote
            # paths and other Apprise-supported sources are supported
            self._prv_keyfile = AppriseAttachment(asset=self.asset)

            # Add our definition to our pgp_key reference
            self._prv_keyfile.add(prv_keyfile)

            # Enforce a private-key-specific size limit; private keys can be
            # larger than public keys (4096-bit RSA, multiple subkeys/UIDs)
            self._prv_keyfile[0].max_file_size = self.max_pgp_private_key_size

        else:
            self._prv_keyfile = None

    def keygen(self, email=None, name=None, force=False):
        """Generates a set of keys based on email configured."""

        try:
            # Create a new RSA key pair with 2048-bit strength
            key = pgpy.PGPKey.new(
                pgpy.constants.PubKeyAlgorithm.RSAEncryptOrSign, 2048
            )

        except NameError:
            # PGPy not installed
            logger.debug("PGPy not installed; keygen disabled")
            return False

        if self._pub_keyfile is not None or not self.path:
            logger.trace(
                "PGP keygen disabled, reason=%s",
                (
                    "keyfile-defined"
                    if self._pub_keyfile is not None
                    else "no-write-path"
                ),
            )
            return False

        if not name:
            name = self.asset.app_id

        if not email:
            email = self.email

        # Prepare our UID
        uid = pgpy.PGPUID.new(name, email=email)

        # Filenames
        file_prefix = email.split("@")[0].lower()

        pub_path = os.path.join(self.path, f"{file_prefix}-pub.asc")
        prv_path = os.path.join(self.path, f"{file_prefix}-prv.asc")

        if os.path.isfile(pub_path) and not force:
            logger.debug(
                "PGP generation skipped; Public Key already exists: %s",
                pub_path,
            )
            return True

        # Persistent Storage Key
        lookup_key = hashlib.sha1(
            os.path.abspath(pub_path).encode("utf-8")
        ).hexdigest()
        if lookup_key in self.__key_lookup:
            # Ensure our key no longer exists
            del self.__key_lookup[lookup_key]

        # Add the user ID to the key
        key.add_uid(
            uid,
            usage={
                pgpy.constants.KeyFlags.Sign,
                pgpy.constants.KeyFlags.EncryptCommunications,
            },
            hashes=[pgpy.constants.HashAlgorithm.SHA256],
            ciphers=[pgpy.constants.SymmetricKeyAlgorithm.AES256],
            compression=[pgpy.constants.CompressionAlgorithm.ZLIB],
        )

        try:
            # Write our keys to disk
            with open(pub_path, "w") as f:
                f.write(str(key.pubkey))

        except OSError as e:
            logger.warning("Error writing PGP file %s", pub_path)
            logger.debug(f"I/O Exception: {e}")

            # Cleanup
            try:
                os.unlink(pub_path)
                logger.trace("Removed %s", pub_path)

            except OSError:
                pass

        try:
            with open(prv_path, "w") as f:
                f.write(str(key))

        except OSError as e:
            logger.warning("Error writing PGP file %s", prv_path)
            logger.debug(f"I/O Exception: {e}")

            try:
                os.unlink(pub_path)
                logger.trace("Removed %s", pub_path)

            except OSError:
                pass

            try:
                os.unlink(prv_path)
                logger.trace("Removed %s", prv_path)

            except OSError:
                pass

            return False

        logger.info(
            "Wrote PGP Keys for %s/%s",
            os.path.dirname(pub_path),
            os.path.basename(pub_path),
        )
        return True

    def _pub_key_candidates(self, *emails):
        """Returns the ordered list of candidate public-key filenames to
        search, highest priority first.

        Shared by public_keyfile() and the diagnostic warning in public_key()
        so both always reflect exactly the same search order.
        """

        # Base candidates, lowest priority first
        fnames = [
            "pgp-public.asc",
            "pgp-pub.asc",
            "public.asc",
            "pub.asc",
        ]

        # Merge the controller's own email with any caller-supplied addresses.
        # Two filenames are prepended per address: localpart shorthand is
        # inserted at index 0 first, then the full address is also inserted
        # at index 0 -- so the full address lands ahead of the localpart in
        # the list (full address has higher priority as it is more specific).
        all_emails = [self.email, *emails] if self.email else list(emails)
        for em in all_emails:
            # Localpart shorthand (e.g. "chris-pub.asc")
            fnames.insert(0, f"{em.split('@')[0].lower()}-pub.asc")
            # Full lowercase email (e.g. "chris@nuxref.com-pub.asc")
            fnames.insert(0, f"{em.lower()}-pub.asc")

        return fnames

    def _prv_key_candidates(self):
        """Returns the ordered list of candidate private-key filenames to
        search, highest priority first.

        Shared by private_keyfile() and the diagnostic warning in private_key()
        so both always reflect exactly the same search order.
        """

        # Base candidates, lowest priority first
        fnames = [
            "pgp-private.asc",
            "pgp-prv.asc",
            "private.asc",
            "prv.asc",
        ]

        # Prefer keys named after the sender's email address when known
        if self.email:
            # Localpart shorthand (e.g. "chris-prv.asc")
            fnames.insert(0, f"{self.email.split('@')[0].lower()}-prv.asc")
            # Full lowercase email (e.g. "chris@nuxref.com-prv.asc")
            fnames.insert(0, f"{self.email.lower()}-prv.asc")

        return fnames

    def public_keyfile(self, *emails):
        """Returns the first match of a useable public key based emails
        provided."""

        if not PGP_SUPPORT:
            msg = "PGP Support unavailable; install PGPy library"
            logger.warning(msg)
            raise ApprisePGPException(msg)

        if self._pub_keyfile is not None:
            # If our code reaches here, then we fetch our public key
            pgp_key = self._pub_keyfile[0]
            if not pgp_key:
                # We could not access the attachment
                logger.error(
                    "Could not access PGP Public Key"
                    f" {pgp_key.url(privacy=True)}."
                )
                return False

            return pgp_key.path

        elif not self.path:
            # No path
            return None

        # Use shared candidate list so the search order is always consistent
        # with any diagnostic messages that reference the same list
        fnames = self._pub_key_candidates(*emails)

        return next(
            (
                os.path.join(self.path, fname)
                for fname in fnames
                if os.path.isfile(os.path.join(self.path, fname))
            ),
            None,
        )

    def private_keyfile(self):
        """Returns the path to the private key file if one can be found.

        Returns the explicit path when prv_keyfile was provided, looks for a
        matching auto-generated key in self.path otherwise.  Returns False
        when an explicit keyfile was given but could not be accessed, and
        None when no key file could be located at all.
        """

        if self._prv_keyfile is not None:
            # Use the explicitly-provided private key attachment
            pgp_key = self._prv_keyfile[0]
            if not pgp_key:
                # We could not access the attachment
                logger.error(
                    "Could not access PGP Private Key"
                    f" {pgp_key.url(privacy=True)}."
                )
                return False

            return pgp_key.path

        elif not self.path:
            # No storage path; cannot search for auto-generated keys
            return None

        # Use shared candidate list so the search order is always consistent
        # with any diagnostic messages that reference the same list
        fnames = self._prv_key_candidates()

        return next(
            (
                os.path.join(self.path, fname)
                for fname in fnames
                if os.path.isfile(os.path.join(self.path, fname))
            ),
            None,
        )

    def private_key(self):
        """Loads and returns the PGP private key object.

        Reads from the explicit prv_keyfile if one was provided, otherwise
        scans the persistent storage path for an auto-generated private key.
        Returns None when no usable private key could be found or loaded.
        Passphrase-protected keys are not supported and will be rejected.
        """

        # Locate the private key file
        path = self.private_keyfile()
        if not path:
            if path is False:
                # An explicit pgpprv= file was given but could not be accessed;
                # keep as WARNING because the user made an explicit choice that
                # failed -- this is always actionable regardless of plugin
                logger.warning("PGP Private Key could not be accessed")

            elif self.path:
                # path is None: storage was searched but nothing matched.
                # Only hash + filename is shown -- never an absolute path.
                ns = os.path.basename(self.path)
                candidates = self._prv_key_candidates()
                shown = ", ".join(f"'{ns}/{fn}'" for fn in candidates[:4])
                extra = ", ..." if len(candidates) > 4 else ""
                logger.debug(
                    "No PGP private key found; searched: %s%s",
                    shown,
                    extra,
                )

            else:
                # No storage path at all -- nothing was searched
                logger.debug("No PGP private key found")

            return None

        # Build a cache key distinct from public-key entries for the same path
        cache_key = hashlib.sha1(
            ("prv:" + os.path.abspath(path)).encode("utf-8")
        ).hexdigest()

        # Return the cached key when it is still valid
        if cache_key in self.__key_lookup:
            entry = self.__key_lookup[cache_key]
            if entry["expires"] > datetime.now(timezone.utc):
                return entry["private_key"]

            # Expired -- remove and re-load below
            del self.__key_lookup[cache_key]

        try:
            with open(path) as key_file:
                private_key, _ = pgpy.PGPKey.from_blob(key_file.read())

        except NameError:
            # PGPy not installed
            logger.debug("PGPy not installed; skipping PGP signing: %s", path)
            return None

        except FileNotFoundError:
            # File was found but disappeared before we could open it
            logger.debug("PGP Private Key file not found: %s", path)
            return None

        except OSError as e:
            logger.warning("Error accessing PGP Private Key file %s", path)
            logger.debug(f"I/O Exception: {e}")
            return None

        except Exception:
            # Malformed or non-PGP file content
            logger.warning(
                "PGP Private Key file could not be parsed: %s", path
            )
            return None

        if private_key.is_protected:
            # Apprise does not support passphrase-protected private keys
            # because it runs unattended; a protected key would block sending
            logger.warning(
                "PGP Private Key is passphrase-protected; "
                "Apprise does not support passphrase-protected keys"
            )
            return None

        # Cache the successfully loaded key
        self.__key_lookup[cache_key] = {
            "private_key": private_key,
            "expires": datetime.now(timezone.utc) + timedelta(seconds=86400),
        }

        return private_key

    def sign(self, message):
        """Creates a detached PGP signature for the given message string.

        Returns a (signature_str, micalg) tuple on success where
        signature_str is the armored PGP signature block and micalg is the
        MIME hash algorithm label (e.g. 'pgp-sha256') for the
        Content-Type header of the multipart/signed container.
        Returns None when signing is not possible.
        """

        # Load our private key
        private_key = self.private_key()
        if not private_key:
            logger.debug("PGP signing skipped: no usable private key")
            return None

        try:
            # Create a detached signature over the message text
            sig = private_key.sign(message)

            # Map the hash algorithm to the MIME micalg label
            micalg = "pgp-" + sig.hash_algorithm.name.lower()

            return (str(sig), micalg)

        except NameError:
            # PGPy not installed; must come before pgpy.errors.PGPError so
            # that evaluating that except clause doesn't itself raise NameError
            logger.debug("PGPy not installed; skipping PGP signing")

        except pgpy.errors.PGPError:
            # Key may lack the Sign capability or be otherwise incompatible
            logger.debug("PGP signing failed; key may lack Sign capability")

        return None

    def _fetch_wkd_key(self, *emails):
        """Attempt a Web Key Directory lookup for each email in turn.

        Returns a pgpy.PGPKey on the first successful fetch, or None if
        WKD is not configured, no emails are provided, or all lookups
        fail.
        """

        if self.wkd is None:
            # No WKD controller configured
            return None

        # Use only the supplied recipient emails when provided. Fall back
        # to self.email only when no recipient emails were supplied
        # (for example, an explicit self-send). Otherwise a missed
        # recipient WKD lookup could incorrectly encrypt to the sender's
        # key, leaving the intended recipient unable to decrypt.
        candidates = (
            list(emails) if emails else ([self.email] if self.email else [])
        )

        for email in candidates:
            if not email:
                continue

            # Derive a stable cache key from the email address
            cache_key = hashlib.sha1(
                ("wkd:" + email.lower()).encode("utf-8")
            ).hexdigest()

            # Return the previously parsed key when still valid
            if cache_key in self.__key_lookup:
                entry = self.__key_lookup[cache_key]
                if entry["expires"] > datetime.now(timezone.utc):
                    return entry["public_key"]
                # Expired; remove so we re-fetch below
                del self.__key_lookup[cache_key]

            # Fetch raw binary key material from WKD
            key_bytes = self.wkd.fetch(email)
            if not key_bytes:
                continue

            try:
                public_key, _ = pgpy.PGPKey.from_blob(key_bytes)

            except NameError:
                # pgpy not installed
                logger.debug(
                    "PGPy not installed; skipping WKD key for %s",
                    email,
                )
                continue

            except Exception:
                # Malformed or unsupported key data from WKD
                logger.debug("WKD key parse failed for %s; skipping", email)
                continue

            # Cache the successfully parsed key
            self.__key_lookup[cache_key] = {
                "public_key": public_key,
                "expires": datetime.now(timezone.utc)
                + timedelta(seconds=86400),
            }

            logger.debug("Loaded PGP public key via WKD for %s", email)
            return public_key

        return None

    def public_key(self, *emails, autogen=None):
        """Opens a spcified pgp public file and returns the key from it which
        is used to encrypt the message."""
        path = self.public_keyfile(*emails)
        if not path:
            # Try Web Key Directory before falling back to autogen
            wkd_key = self._fetch_wkd_key(*emails)
            if wkd_key is not None:
                return wkd_key

            if (
                autogen if autogen is not None else self.asset.pgp_autogen
            ) and self.keygen(*emails):
                path = self.public_keyfile(*emails)
                if path:
                    # We should get a hit now
                    return self.public_key(*emails)

            # All discovery methods exhausted (local file, WKD, autogen).
            # Log at DEBUG so the caller's warning (with plugin-specific hints)
            # is the only user-visible message.  Only the namespace hash +
            # filename is shown -- never an absolute path -- so no sensitive
            # filesystem layout is revealed.
            if self.path:
                ns = os.path.basename(self.path)
                candidates = self._pub_key_candidates(*emails)
                shown = ", ".join(f"'{ns}/{fn}'" for fn in candidates[:4])
                extra = ", ..." if len(candidates) > 4 else ""
                logger.debug(
                    "No PGP public key found; searched: %s%s",
                    shown,
                    extra,
                )
            else:
                logger.debug("No PGP public key found")

            return None

        # Persistent Storage Key
        key = hashlib.sha1(os.path.abspath(path).encode("utf-8")).hexdigest()
        if key in self.__key_lookup:
            # Take an early exit
            return self.__key_lookup[key]["public_key"]

        try:
            with open(path) as key_file:
                public_key, _ = pgpy.PGPKey.from_blob(key_file.read())

        except NameError:
            # PGPy not installed
            logger.debug("PGPy not installed; skipping PGP support: %s", path)
            return None

        except FileNotFoundError:
            # Generate keys
            logger.debug("PGP Public Key file not found: %s", path)
            return None

        except OSError as e:
            logger.warning("Error accessing PGP Public Key file %s", path)
            logger.debug(f"I/O Exception: {e}")
            return None

        except Exception:
            # Malformed or non-PGP file content (e.g. pgpy.errors.PGPError)
            logger.warning("PGP Public Key file could not be parsed: %s", path)
            return None

        self.__key_lookup[key] = {
            "public_key": public_key,
            "expires": datetime.now(timezone.utc) + timedelta(seconds=86400),
        }
        return public_key

    # Encrypt message using the recipient's public key
    def encrypt(self, message, *emails, autogen=None):
        """If provided a path to a pgp-key, content is encrypted.

        Pass autogen=False to suppress key auto-generation during the
        public-key lookup.  This is used in sign mode for opportunistic
        encryption: only encrypt when a pre-existing key is found.
        """

        # Acquire our key; autogen controls whether a missing key is created
        public_key = self.public_key(*emails, autogen=autogen)
        if not public_key:
            # Encryption not possible
            return False

        try:
            message_object = pgpy.PGPMessage.new(message)
            encrypted_message = public_key.encrypt(message_object)
            return str(encrypted_message)

        except pgpy.errors.PGPError:
            # Encryption not Possible
            logger.debug("PGP Public Key Corruption; encryption not possible")

        except NameError:
            # PGPy not installed
            logger.debug("PGPy not installed; Skipping PGP encryption")

        return None

    def prune(self):
        """Prunes old entries from the public_key index."""
        self.__key_lookup = {
            key: value
            for key, value in self.__key_lookup.items()
            if value["expires"] > datetime.now(timezone.utc)
        }

        # Also prune the WKD in-memory cache when a controller is set
        if self.wkd is not None:
            self.wkd.prune()

    @property
    def pub_keyfile(self):
        """Returns the Public Keyfile Path if set otherwise it returns None
        This property returns False if a keyfile was provided, but was
        invalid."""
        return (
            None
            if self._pub_keyfile is None
            else (
                False
                if not self._pub_keyfile[0]
                else self._pub_keyfile[0].path
            )
        )

    @property
    def prv_keyfile(self):
        """Returns the Private Keyfile Path if set, otherwise returns None.
        Returns False when an explicit keyfile was provided but could not
        be accessed."""
        return (
            None
            if self._prv_keyfile is None
            else (
                False
                if not self._prv_keyfile[0]
                else self._prv_keyfile[0].path
            )
        )



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Sign-up at https://ringcentral.com
#
# Create a server/bot app in the developer portal:
#  https://developers.ringcentral.com/
#
# Steps:
#  1. Log in and click "Create App".
#  2. Choose "REST API App" -> "Server/Bot (No UI)".
#  3. Under permissions enable "SMS" and "MMS" (MMS is needed for
#     attachment support -- Apprise selects MMS automatically when
#     attachments are present).
#  4. Copy the Client ID and Client Secret from the app credentials tab.
#
# Two authentication modes are supported:
#
# BASIC (password) mode -- use a RingCentral user password:
#   ringc://SourcePhoneNo:Password@ClientID/ClientSecret
#   ringc://SourcePhoneNo:Password@ClientID/ClientSecret/ToPhoneNo
#   ringc://SourcePhoneNo:Password@ClientID/ClientSecret/To1/To2/ToN
#
# JWT mode -- use a JWT token generated in the developer portal:
#   ringc://SourcePhoneNo:JWTToken@ClientID/ClientSecret
#   ringc://SourcePhoneNo:JWTToken@ClientID/ClientSecret/ToPhoneNo
#
# Alternatively, supply credentials as query parameters:
#   ringc://_?token=JWT&secret=ClientSecret&from=SourcePhoneNo
#
# When attachments are provided, Apprise automatically uses the MMS
# endpoint; otherwise SMS is used.
#
# API references:
# - https://developers.ringcentral.com/api-reference/SMS/createSMSMessage
# - https://developers.ringcentral.com/api-reference/MMS/createMMSMessage
# - https://developers.ringcentral.com/api-reference/OAuth-2.0/getToken

import base64
import contextlib
from json import dumps, loads
from time import time

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import is_phone_no, parse_phone_no, validate_regex
from .base import NotifyBase


class RingCentralAuthMode:
    """Authentication modes supported by RingCentral."""

    # Username + password credential flow
    BASIC = "basic"

    # JWT bearer token credential flow
    JWT = "jwt"


# Valid auth mode choices
RINGCENTRAL_AUTH_MODES = (
    RingCentralAuthMode.BASIC,
    RingCentralAuthMode.JWT,
)


class RingCentralEnvironment:
    """RingCentral API environment targets."""

    # Live production environment
    PRODUCTION = "prod"

    # Sandbox / devtest environment
    SANDBOX = "sandbox"


# Valid environment choices
RINGCENTRAL_ENVIRONMENTS = (
    RingCentralEnvironment.PRODUCTION,
    RingCentralEnvironment.SANDBOX,
)

# Maps environment choice to the URL infix used by the API
RINGCENTRAL_ENV_URL_SUFFIX = {
    RingCentralEnvironment.PRODUCTION: "",
    RingCentralEnvironment.SANDBOX: ".devtest",
}


class NotifyRingCentral(NotifyBase):
    """A wrapper for RingCentral Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "RingCentral"

    # The services URL
    service_url = "https://ringcentral.com/"

    # The default protocol
    secure_protocol = "ringc"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://appriseit.com/services/ringcentral/"

    # Attachments are supported; MMS is selected automatically when present
    attachment_support = True

    # SMS endpoint (environment is substituted at send time)
    notify_url_sms = (
        "https://platform{environment}.ringcentral.com/"
        "restapi/v1.0/account/~/extension/~/sms"
    )

    # MMS endpoint (used automatically when attachments are present)
    notify_url_mms = (
        "https://platform{environment}.ringcentral.com/"
        "restapi/v1.0/account/~/extension/~/mms"
    )

    # OAuth token endpoint
    access_token_url = (
        "https://platform{environment}.ringcentral.com/restapi/oauth/token"
    )

    # Token revocation endpoint
    revoke_token_url = (
        "https://platform{environment}.ringcentral.com/restapi/oauth/revoke"
    )

    # Access token lifetime in seconds (60 minutes)
    access_token_ttl = 3600

    # Refresh token lifetime in seconds (1 week)
    refresh_token_ttl = 604800

    # SMS body length limit
    body_maxlen = 160

    # Titles are not supported; prepend to body instead
    title_maxlen = 0

    # Define object URL templates
    templates = (
        # source:token@client_id/secret (no targets -- loopback to source)
        "{schema}://{from_phone}:{token}@{client_id}/{secret}/",
        # source:token@client_id/secret/target[/target...]
        "{schema}://{from_phone}:{token}@{client_id}/{secret}/{targets}",
        # token@client_id/secret/source (source in path, JWT/password first)
        "{schema}://{token}@{client_id}/{secret}/{from_phone}",
        # token@client_id/secret/source/target[/target...]
        "{schema}://{token}@{client_id}/{secret}/{from_phone}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "token": {
                "name": _("Token / Password"),
                "type": "string",
                "required": True,
                "private": True,
            },
            "client_id": {
                "name": _("Client ID"),
                "type": "string",
                "required": True,
                "regex": (r"^[a-z0-9_-]+$", "i"),
                "private": True,
            },
            "secret": {
                "name": _("Client Secret"),
                "type": "string",
                "required": True,
                "regex": (r"^[a-z0-9_-]+$", "i"),
                "private": True,
                "map_to": "client_secret",
            },
            "from_phone": {
                "name": _("From Phone No"),
                "type": "string",
                "required": True,
                "map_to": "source",
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "alias_of": "from_phone",
            },
            "source": {
                "alias_of": "from_phone",
            },
            "env": {
                "name": _("Environment"),
                "type": "choice:string",
                "values": RINGCENTRAL_ENVIRONMENTS,
                "default": RingCentralEnvironment.PRODUCTION,
                "map_to": "environment",
            },
            "token": {
                "alias_of": "token",
            },
            "secret": {
                "alias_of": "secret",
            },
            "mode": {
                "name": _("Authentication Mode"),
                "type": "choice:string",
                "values": RINGCENTRAL_AUTH_MODES,
                "map_to": "mode",
            },
        },
    )

    def __init__(
        self,
        source,
        targets=None,
        environment=None,
        token=None,
        client_id=None,
        client_secret=None,
        mode=None,
        **kwargs,
    ):
        """Initialize RingCentral Object."""
        super().__init__(**kwargs)

        # Internal OAuth state
        self._access_token = None
        self._expire_time = 0.0
        self._scope = None
        self._owner = None
        self._endpoint_id = None

        # Resolve authentication mode
        if isinstance(mode, str):
            _mode = mode.lower().strip()
            match = (
                next(
                    (m for m in RINGCENTRAL_AUTH_MODES if m.startswith(_mode)),
                    None,
                )
                if _mode
                else None
            )

            if not match:
                msg = (
                    "An invalid RingCentral Authentication Mode "
                    "({}) was specified.".format(mode)
                )
                self.logger.warning(msg)
                raise TypeError(msg)

            # Store resolved mode
            self.mode = match

        else:
            # Default to BASIC (password) auth
            self.mode = RingCentralAuthMode.BASIC

        # Validate Client ID
        self.client_id = validate_regex(
            client_id, *self.template_tokens["client_id"]["regex"]
        )
        if not self.client_id:
            msg = (
                "An invalid RingCentral Client ID ({}) was specified.".format(
                    client_id
                )
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Validate Client Secret
        self.client_secret = validate_regex(
            client_secret, *self.template_tokens["secret"]["regex"]
        )
        if not self.client_secret:
            msg = (
                "An invalid RingCentral Client Secret "
                "({}) was specified.".format(client_secret)
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Validate JWT token (only strict regex check in JWT mode)
        if self.mode == RingCentralAuthMode.JWT:
            self.token = validate_regex(token, r"^[a-z0-9._-]+$", "i")
            if not self.token:
                msg = (
                    "An invalid RingCentral JWT Token "
                    "({}) was specified.".format(token)
                )
                self.logger.warning(msg)
                raise TypeError(msg)
        else:
            # BASIC mode -- token holds the user password (not validated)
            self.token = token

        # Validate source (sender) phone number
        result = is_phone_no(source)
        if not result:
            msg = (
                "The RingCentral source (From) phone # "
                "({}) is invalid.".format(source)
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store normalised source digits only
        self.source = result["full"]

        # Resolve environment
        _environment = (
            environment.lower().strip()
            if isinstance(environment, str)
            else RingCentralEnvironment.PRODUCTION
        )
        match = (
            next(
                (
                    e
                    for e in RINGCENTRAL_ENVIRONMENTS
                    if e.startswith(_environment)
                ),
                None,
            )
            if _environment
            else None
        )
        if not match:
            msg = (
                "An invalid RingCentral environment "
                "({}) was specified.".format(environment)
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store resolved environment
        self.environment = match

        # Parse target phone numbers, dropping invalid ones
        self.targets = []
        for target in parse_phone_no(targets):
            result = is_phone_no(target)
            if not result:
                self.logger.warning(
                    "Dropped invalid RingCentral phone # (%s) specified.",
                    target,
                )
                continue

            # Store normalised digits
            self.targets.append(result["full"])

        return

    def login(self):
        """Authenticate with the RingCentral OAuth token endpoint."""

        if self._expire_time >= time():
            # Already authenticated
            return True

        # Build the token request URL
        url = self.access_token_url.format(
            environment=RINGCENTRAL_ENV_URL_SUFFIX[self.environment],
        )

        # Encode client credentials for Basic auth header
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": "Basic {}".format(
                str(
                    base64.b64encode(
                        bytes(
                            self.client_id + ":" + self.client_secret,
                            "utf8",
                        )
                    ),
                    "utf8",
                )
            ),
        }

        # Clear any stale token state
        self._access_token = None
        self._scope = None
        self._owner = None
        self._endpoint_id = None

        # Build grant payload for the selected auth mode
        if self.mode == RingCentralAuthMode.JWT:
            payload = {
                "grant_type": ("urn:ietf:params:oauth:grant-type:jwt-bearer"),
                "assertion": self.token,
            }
        else:
            # BASIC password grant
            payload = {
                "grant_type": "password",
                "username": "+" + self.source,
                "password": self.token,
                "access_token_ttl": self.access_token_ttl,
                "refresh_token_ttl": self.refresh_token_ttl,
            }

        # Send authentication request (no throttle for auth calls)
        status, response = self._send(
            url,
            payload,
            headers,
            name="auth.login",
            throttle=False,
        )

        if status:
            # Store token and expiry
            self._access_token = response.get("access_token")
            self._expire_time = time() + response.get("expires_in", 0)
            self._scope = response.get("scope")
            self._owner = response.get("owner_id")
            self._endpoint_id = response.get("endpoint_id")

        # A 200 OK without an access_token is not a usable login
        return bool(self._access_token)

    def logout(self):
        """Revoke the current access token."""

        if not self._access_token or self._expire_time < time():
            # Nothing to revoke
            return

        # Build the revocation URL
        url = self.revoke_token_url.format(
            environment=RINGCENTRAL_ENV_URL_SUFFIX[self.environment],
        )

        # Encode client credentials
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": "Basic {}".format(
                str(
                    base64.b64encode(
                        bytes(
                            self.client_id + ":" + self.client_secret,
                            "utf8",
                        )
                    ),
                    "utf8",
                )
            ),
        }

        # Revocation payload
        payload = {"token": self._access_token}

        # Send revocation request (no throttle for auth calls)
        self._send(
            url,
            payload,
            headers,
            name="auth.logout",
            throttle=False,
        )

        # Clear token state
        self._access_token = None
        self._expire_time = 0.0
        self._scope = None
        self._owner = None
        self._endpoint_id = None

        return

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform RingCentral Notification."""

        # Error tracking across multiple targets
        has_error = False

        # Authenticate before sending
        if not self.login():
            self.logger.warning(
                "RingCentral %s authentication failed.", self.mode
            )
            return False

        # Bearer auth header token (reused across send calls)
        auth_header = "Bearer {}".format(self._access_token)

        # Auto-select MMS when attachments are present, SMS otherwise
        use_mms = bool(attach and self.attachment_support and len(attach))

        # Build the endpoint URL for the selected message type
        if use_mms:
            notify_url = self.notify_url_mms.format(
                environment=RINGCENTRAL_ENV_URL_SUFFIX[self.environment],
            )
        else:
            notify_url = self.notify_url_sms.format(
                environment=RINGCENTRAL_ENV_URL_SUFFIX[self.environment],
            )

        # If no targets specified, send to own number (loopback test)
        targets = list(self.targets) if self.targets else [self.source]

        for target in targets:
            # Message metadata for this recipient
            metadata = {
                "from": {"phoneNumber": "+" + self.source},
                "to": [{"phoneNumber": "+" + target}],
                "text": body,
            }

            if use_mms:
                # MMS: multipart/form-data -- JSON metadata + attachments
                # Content-Type is set automatically by requests for multipart
                headers = {
                    "Authorization": auth_header,
                }
                files = [
                    (
                        "json",
                        (None, dumps(metadata), "application/json"),
                    ),
                ]

                # Track opened handles for guaranteed cleanup
                handles = []
                attach_ok = True

                try:
                    # Build attachment parts; abort target on first failure
                    for attachment in attach:
                        # Verify the attachment is accessible
                        if not attachment:
                            self.logger.warning(
                                "Could not access RingCentral attachment %s.",
                                attachment.url(privacy=True),
                            )
                            attach_ok = False
                            break

                        # Open handle; guard against I/O errors
                        try:
                            handle = attachment.open()
                        except OSError as exc:
                            self.logger.warning(
                                "An I/O error occurred reading "
                                "RingCentral attachment %s.",
                                attachment.name,
                            )
                            self.logger.debug("I/O Exception: %s", str(exc))
                            attach_ok = False
                            break

                        # Register handle for cleanup
                        handles.append(handle)
                        # Append the attachment part
                        files.append(
                            (
                                "attachment",
                                (
                                    attachment.name,
                                    handle,
                                    attachment.mimetype,
                                ),
                            )
                        )

                    if not attach_ok:
                        # Skip this target; mark overall failure
                        has_error = True
                        continue

                    # Send MMS notification (multipart)
                    status, _ = self._send(
                        notify_url, None, headers, files=files
                    )

                finally:
                    # Close all handles whether we succeeded or failed
                    for handle in handles:
                        handle.close()

            else:
                # SMS: plain JSON payload
                headers = {
                    "Content-Type": "application/json",
                    "Authorization": auth_header,
                }
                status, _ = self._send(notify_url, dumps(metadata), headers)

            if status:
                self.logger.info(
                    "Sent RingCentral notification to %s.", target
                )
            else:
                # Mark failure and continue to remaining targets
                has_error = True

        return not has_error

    def _send(
        self,
        url,
        payload,
        headers,
        name="notification",
        throttle=True,
        files=None,
    ):
        """POST helper shared by login, logout, and send calls."""

        headers.update(
            {
                # Minimum required headers
                "User-Agent": self.app_id,
                "Accept": "application/json",
            }
        )

        self.logger.debug(
            "RingCentral POST URL: %s (cert_verify=%s)",
            url,
            self.verify_certificate,
        )
        self.logger.debug("RingCentral Payload: %s", payload)

        if throttle:
            # Throttle notification calls; auth calls skip this
            self.throttle()

        content = None
        try:
            r = requests.post(
                url,
                data=payload,
                headers=headers,
                files=files,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
                allow_redirects=self.redirects,
            )
            self.logger.trace("RingCentral Response: %s", r.content)

            try:
                content = loads(r.content)

            except (AttributeError, TypeError, ValueError):
                # ValueError = r.content is unparsable
                # TypeError = r.content is None
                # AttributeError = r is None
                content = {}

            if r.status_code != requests.codes.ok:
                # We had a failure
                status_str = NotifyRingCentral.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send RingCentral %s: %s%serror=%s.",
                    name,
                    status_str,
                    ", " if status_str else "",
                    r.status_code,
                )
                self.logger.debug("Response Details:\r\n%s", r.content)

            else:
                # We were successful
                return (True, content)

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending RingCentral %s.",
                name,
            )
            self.logger.debug("Socket Exception: %s", str(e))

        return (False, content)

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another similar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol,
            self.client_id,
            self.client_secret,
            self.token,
            self.source,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "env": str(self.environment),
            "mode": str(self.mode),
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return (
            "{schema}://{source}:{token}@{client_id}"
            "/{client_secret}/{targets}/?{params}".format(
                schema=self.secure_protocol,
                source=NotifyRingCentral.quote(self.source, safe=""),
                token=self.pprint(
                    self.token,
                    privacy,
                    mode=(
                        PrivacyMode.Secret
                        if self.mode == RingCentralAuthMode.BASIC
                        else PrivacyMode.Outer
                    ),
                    safe="",
                ),
                client_id=self.pprint(self.client_id, privacy, safe=""),
                client_secret=self.pprint(
                    self.client_secret,
                    privacy,
                    mode=PrivacyMode.Secret,
                    safe="",
                ),
                targets="/".join(
                    NotifyRingCentral.quote(x, safe="") for x in self.targets
                ),
                params=NotifyRingCentral.urlencode(params),
            )
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.targets) if self.targets else 1

    def __del__(self):
        """Destructor -- revoke the auth token on cleanup."""
        with contextlib.suppress(Exception):
            self.logout()

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow
        us to re-instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Path elements: first segment is client_secret, rest are targets
        results["targets"] = NotifyRingCentral.split_path(results["fullpath"])

        # Host is the client_id
        results["client_id"] = NotifyRingCentral.unquote(results["host"])

        # First path segment is the client_secret
        results["client_secret"] = (
            results["targets"].pop(0) if results["targets"] else None
        )

        # Determine source phone and token from URL user[:password] form
        if not results.get("password"):
            # ringc://token@client_id/secret/source_phone[/targets]
            results["source"] = (
                results["targets"].pop(0) if results["targets"] else None
            )
            results["token"] = NotifyRingCentral.unquote(results["user"])
        else:
            # ringc://source_phone:token@client_id/secret[/targets]
            results["source"] = NotifyRingCentral.unquote(results["user"])
            results["token"] = NotifyRingCentral.unquote(results["password"])

        # Environment (from ?env= query parameter)
        if "env" in results["qsd"] and results["qsd"]["env"]:
            results["environment"] = NotifyRingCentral.unquote(
                results["qsd"]["env"]
            )

        # ?token= overrides the token extracted from the URL path;
        # must be applied before mode auto-detection so that a JWT
        # supplied via query parameter is detected correctly
        if "token" in results["qsd"] and results["qsd"]["token"]:
            results["token"] = NotifyRingCentral.unquote(
                results["qsd"]["token"]
            )

        # Auth mode: ?mode= wins; otherwise auto-detect from token length
        if "mode" in results["qsd"] and results["qsd"]["mode"]:
            results["mode"] = NotifyRingCentral.unquote(results["qsd"]["mode"])
        elif results.get("token") and len(results["token"]) > 60:
            # JWT tokens are significantly longer than passwords
            results["mode"] = RingCentralAuthMode.JWT
        else:
            results["mode"] = RingCentralAuthMode.BASIC

        # ?secret= overrides the client secret extracted from the path
        if "secret" in results["qsd"] and results["qsd"]["secret"]:
            results["client_secret"] = NotifyRingCentral.unquote(
                results["qsd"]["secret"]
            )

        # ?from= / ?source= override source phone number
        if "from" in results["qsd"] and results["qsd"]["from"]:
            results["source"] = NotifyRingCentral.unquote(
                results["qsd"]["from"]
            )
        if "source" in results["qsd"] and results["qsd"]["source"]:
            results["source"] = NotifyRingCentral.unquote(
                results["qsd"]["source"]
            )

        # ?to= appends additional target phone numbers
        if "to" in results["qsd"] and results["qsd"]["to"]:
            results["targets"] += NotifyRingCentral.parse_phone_no(
                results["qsd"]["to"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Amazon Chime Incoming Webhooks
#
# To create an incoming webhook for an Amazon Chime chat room:
#  1. Open Amazon Chime and navigate to the chat room you want to
#     send notifications to.
#  2. Choose the gear icon in the top-right corner of the chat room.
#  3. Choose "Manage webhooks and bots".
#  4. Choose "Add webhook", give it a name, then choose "Create".
#  5. To copy the webhook URL, choose "Copy URL" next to your new
#     webhook in the list.
#
# The webhook URL will look something like:
#   https://hooks.chime.aws/incomingwebhooks/
#       xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx?token=AaBbCcDd%3D%3D
#
# Where the path segment after /incomingwebhooks/ is the Webhook ID
# and the ?token= query parameter is the authentication token.
#
# Your Apprise URL would be assembled as:
#   chime://{WebhookID}/{Token}
#
# For example:
#   chime://xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/AaBbCcDd%3D%3D
#
# The token value may include base64 padding characters (=). Apprise
# handles encoding automatically; you can paste the token exactly as
# it appears in the Chime webhook URL (already URL-decoded form) or
# as-is from the ?token= query parameter.
#
# References:
# - https://docs.aws.amazon.com/chime/latest/ug/webhooks.html
# - https://docs.aws.amazon.com/chime/latest/APIReference/\
#       API_CreateIncomingWebhook.html

from json import dumps
import re

import requests

from ..common import NotifyFormat, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import validate_regex
from .base import NotifyBase

# Extend HTTP Error Messages
CHIME_HTTP_ERROR_MAP = {
    401: "Unauthorized - Invalid Webhook Token.",
    403: "Forbidden - Webhook has been disabled or revoked.",
    404: "Not Found - Invalid Webhook ID.",
    429: "Too many requests; rate-limit exceeded.",
}

# Amazon Chime Incoming Webhook base URL
CHIME_WEBHOOK_URL = (
    "https://hooks.chime.aws/incomingwebhooks/{webhook_id}?token={token}"
)

# Validates a Chime webhook ID (UUID-like hex string with hyphens,
# but we keep this flexible to handle any future format changes)
IS_WEBHOOK_ID = re.compile(r"^[a-z0-9]([a-z0-9-]*[a-z0-9])?$", re.I)


class NotifyChime(NotifyBase):
    """A wrapper for Amazon Chime Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Amazon Chime"

    # The services URL
    service_url = "https://aws.amazon.com/chime/"

    # The default secure protocol
    secure_protocol = "chime"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://appriseit.com/services/chime/"

    # Amazon Chime Webhook URL (used in send())
    notify_url = CHIME_WEBHOOK_URL

    # Amazon Chime webhooks support Markdown in the Content field
    notify_format = NotifyFormat.MARKDOWN

    # Chime webhooks have no native title field; title_maxlen = 0
    # causes the framework to prepend the title to the body automatically
    title_maxlen = 0

    # Maximum allowed characters in the Content field
    body_maxlen = 4096

    # Chime incoming webhooks do not support file attachments;
    # attachments can only be sent through the Chime SDK/API
    attachment_support = False

    # Define object URL templates
    templates = ("{schema}://{webhook_id}/{token}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "webhook_id": {
                "name": _("Webhook ID"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "token": {
                "name": _("Webhook Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "webhook_id": {
                "alias_of": "webhook_id",
            },
            "token": {
                "alias_of": "token",
            },
        },
    )

    def __init__(self, webhook_id, token, **kwargs):
        """Initialize Amazon Chime Object."""
        super().__init__(**kwargs)

        # Validate the Webhook ID (any non-empty, non-whitespace string)
        self.webhook_id = validate_regex(webhook_id)
        if not self.webhook_id:
            msg = (
                "An invalid Amazon Chime Webhook ID "
                "({}) was specified.".format(webhook_id)
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Validate the Webhook Token
        self.token = validate_regex(token)
        if not self.token:
            msg = (
                "An invalid Amazon Chime Webhook Token "
                "({}) was specified.".format(token)
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Amazon Chime Notification."""

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json; charset=utf-8",
        }

        # Prepare our payload; Chime webhooks accept a single Content field
        payload = {
            "Content": body,
        }

        # Construct the full webhook URL; the token must be URL-encoded
        # because it may contain base64 padding characters (=, +, /)
        notify_url = self.notify_url.format(
            webhook_id=self.webhook_id,
            token=NotifyChime.quote(self.token, safe=""),
        )

        self.logger.debug(
            "Amazon Chime POST URL: %s (cert_verify=%s)",
            notify_url,
            self.verify_certificate,
        )
        self.logger.debug("Amazon Chime Payload: %s", str(payload))

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                notify_url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
                allow_redirects=self.redirects,
            )

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyChime.http_response_code_lookup(
                    r.status_code, CHIME_HTTP_ERROR_MAP
                )

                self.logger.warning(
                    "Failed to send Amazon Chime notification: "
                    "{}{}error={}.".format(
                        status_str,
                        ", " if status_str else "",
                        r.status_code,
                    )
                )

                self.logger.debug(
                    "Response Details:\r\n%r", (r.content or b"")[:2000]
                )

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Amazon Chime notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred posting to Amazon Chime."
            )
            self.logger.debug("Socket Exception: %s", str(e))
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol,
            self.webhook_id,
            self.token,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Acquire any global URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        return "{schema}://{webhook_id}/{token}/?{params}".format(
            schema=self.secure_protocol,
            webhook_id=self.pprint(self.webhook_id, privacy, safe=""),
            token=self.pprint(self.token, privacy, safe=""),
            params=NotifyChime.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow
        us to re-instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Webhook ID is in the host position
        results["webhook_id"] = NotifyChime.unquote(results["host"])

        # The token is the first entry in the URL path
        tokens = NotifyChime.split_path(results["fullpath"])
        results["token"] = tokens.pop(0) if tokens else None

        # Allow ?token= to override the path-supplied token
        if "token" in results["qsd"] and results["qsd"]["token"]:
            results["token"] = NotifyChime.unquote(results["qsd"]["token"])

        # Allow ?webhook_id= to override the host-supplied webhook ID
        if "webhook_id" in results["qsd"] and results["qsd"]["webhook_id"]:
            results["webhook_id"] = NotifyChime.unquote(
                results["qsd"]["webhook_id"]
            )

        return results

    @staticmethod
    def parse_native_url(url):
        """Support native Amazon Chime webhook URLs.

        For example:
          https://hooks.chime.aws/incomingwebhooks/
              xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx?token=AaBbCcDd%3D%3D
        """
        result = re.match(
            r"^https?://hooks\.chime\.aws/incomingwebhooks/"
            r"(?P<webhook_id>[a-z0-9][a-z0-9-]*[a-z0-9])/?"
            r"(\?.*token=(?P<token>[^&#]+).*)?$",
            url,
            re.I,
        )
        if result:
            # Unquote the token from the native URL's query string so
            # parse_url() receives the decoded form and can store it
            token = result.group("token") or ""
            return NotifyChime.parse_url(
                "{schema}://{webhook_id}/{token}".format(
                    schema=NotifyChime.secure_protocol,
                    webhook_id=result.group("webhook_id"),
                    # Re-encode so parse_url / split_path decode it once
                    token=NotifyChime.quote(
                        NotifyChime.unquote(token), safe=""
                    ),
                )
            )

        return None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Disable logging for a cleaner testing output
import logging
import os
import sys
from unittest import mock

import pytest

from apprise.utils import pgp as pgp_module
from apprise.utils.pgp import ApprisePGPController, _ensure_imghdr_shim

logging.disable(logging.CRITICAL)

# Path to the pre-generated private key fixture used by signing tests
_VAR_DIR = os.path.join(os.path.dirname(__file__), "var", "pgp")
_VALID_PRV_ASC = os.path.join(_VAR_DIR, "valid-prv.asc")


# ---------------------------------------------------------------------------
# imghdr shim
# ---------------------------------------------------------------------------


def test_imghdr_shim_when_present():
    """_ensure_imghdr_shim() leaves sys.modules intact when imghdr imports."""

    # Guarantee imghdr is importable for this test
    real_imghdr = sys.modules.get("imghdr")
    try:
        import imghdr as _real

        # If we got here, imghdr is available on this Python version
        _ensure_imghdr_shim()

        # sys.modules entry must not have been replaced by our shim
        assert sys.modules.get("imghdr") is _real

    except ImportError:
        # Python 3.13+ -- imghdr is gone; skip this branch
        pytest.skip("imghdr not present on this Python version")

    finally:
        # Restore whatever was there before
        if real_imghdr is None:
            sys.modules.pop("imghdr", None)
        else:
            sys.modules["imghdr"] = real_imghdr


def test_imghdr_shim_when_absent():
    """_ensure_imghdr_shim() installs a working shim when imghdr is missing."""

    # Remove any real imghdr from sys.modules and make import fail
    saved = sys.modules.pop("imghdr", None)
    try:
        # Setting the key to None causes 'import imghdr' to raise ImportError
        sys.modules["imghdr"] = None

        _ensure_imghdr_shim()

        shim = sys.modules.get("imghdr")
        assert shim is not None
        assert callable(shim.what)

        # The shim must return None (safe fallback for all inputs)
        assert shim.what() is None
        assert shim.what(file=None, h=b"\xff\xd8\xff") is None
        assert shim.what(None, h=b"some data") is None

    finally:
        # Restore original state
        if saved is None:
            sys.modules.pop("imghdr", None)
        else:
            sys.modules["imghdr"] = saved


# ---------------------------------------------------------------------------
# ApprisePGPController -- WKD integration
# ---------------------------------------------------------------------------


def test_pgp_controller_wkd_none_by_default(tmpdir):
    """ApprisePGPController has wkd=None unless explicitly provided."""
    ctrl = ApprisePGPController(path=str(tmpdir))
    assert ctrl.wkd is None


def test_pgp_controller_accepts_wkd(tmpdir):
    """ApprisePGPController stores the WKD controller it is given."""
    mock_wkd = mock.Mock()
    ctrl = ApprisePGPController(path=str(tmpdir), wkd=mock_wkd)
    assert ctrl.wkd is mock_wkd


def test_fetch_wkd_key_no_controller(tmpdir):
    """_fetch_wkd_key() returns None when no WKD controller is set."""
    ctrl = ApprisePGPController(path=str(tmpdir))
    result = ctrl._fetch_wkd_key("user@example.com")
    assert result is None


def test_fetch_wkd_key_empty_emails(tmpdir):
    """_fetch_wkd_key() returns None when no email candidates exist."""
    mock_wkd = mock.Mock()
    ctrl = ApprisePGPController(path=str(tmpdir), wkd=mock_wkd)
    # No self.email set and no positional emails
    result = ctrl._fetch_wkd_key()
    assert result is None
    mock_wkd.fetch.assert_not_called()


def test_fetch_wkd_key_skips_none_in_candidate_list(tmpdir):
    """_fetch_wkd_key() skips None/empty entries in the email candidates."""
    mock_wkd = mock.Mock()
    mock_wkd.fetch.return_value = None
    ctrl = ApprisePGPController(path=str(tmpdir), wkd=mock_wkd)
    # None is a valid positional argument; it must be skipped, not crash
    result = ctrl._fetch_wkd_key(None, "user@example.com")
    assert result is None
    # fetch must only be called for the valid address, not for None
    mock_wkd.fetch.assert_called_once_with("user@example.com")


def test_fetch_wkd_key_wkd_returns_none(tmpdir):
    """_fetch_wkd_key() returns None when WKD fetch yields nothing."""
    mock_wkd = mock.Mock()
    mock_wkd.fetch.return_value = None
    ctrl = ApprisePGPController(path=str(tmpdir), wkd=mock_wkd)
    result = ctrl._fetch_wkd_key("user@example.com")
    assert result is None


@pytest.mark.skipif("pgpy" not in sys.modules, reason="Requires PGPy")
def test_fetch_wkd_key_success(tmpdir):
    """_fetch_wkd_key() parses and caches a key returned by WKD."""
    import pgpy

    # Generate a fresh key pair to use as our fake WKD payload
    key = pgpy.PGPKey.new(
        pgpy.constants.PubKeyAlgorithm.RSAEncryptOrSign, 2048
    )
    uid = pgpy.PGPUID.new("Test", email="user@example.com")
    key.add_uid(
        uid,
        usage={
            pgpy.constants.KeyFlags.Sign,
            pgpy.constants.KeyFlags.EncryptCommunications,
        },
        hashes=[pgpy.constants.HashAlgorithm.SHA256],
        ciphers=[pgpy.constants.SymmetricKeyAlgorithm.AES256],
        compression=[pgpy.constants.CompressionAlgorithm.ZLIB],
    )
    pub_bytes = str(key.pubkey).encode()

    mock_wkd = mock.Mock()
    mock_wkd.fetch.return_value = pub_bytes
    ctrl = ApprisePGPController(path=str(tmpdir), wkd=mock_wkd)

    result = ctrl._fetch_wkd_key("user@example.com")
    assert result is not None
    assert isinstance(result, pgpy.PGPKey)
    mock_wkd.fetch.assert_called_once_with("user@example.com")

    # Second call must use the parsed-key cache -- wkd.fetch must NOT be
    # called again (the cache check now happens before wkd.fetch())
    mock_wkd.fetch.reset_mock()
    result2 = ctrl._fetch_wkd_key("user@example.com")
    assert result2 is not None
    mock_wkd.fetch.assert_not_called()


@pytest.mark.skipif("pgpy" not in sys.modules, reason="Requires PGPy")
def test_fetch_wkd_key_expired_cache_refetches(tmpdir):
    """_fetch_wkd_key() drops an expired parsed-key cache entry and
    re-fetches from WKD rather than returning the stale key."""
    from datetime import datetime, timedelta, timezone
    import hashlib

    import pgpy

    # Generate a key to serve as the fresh WKD result
    key = pgpy.PGPKey.new(
        pgpy.constants.PubKeyAlgorithm.RSAEncryptOrSign, 2048
    )
    uid = pgpy.PGPUID.new("Fresh", email="user@example.com")
    key.add_uid(
        uid,
        usage={
            pgpy.constants.KeyFlags.Sign,
            pgpy.constants.KeyFlags.EncryptCommunications,
        },
        hashes=[pgpy.constants.HashAlgorithm.SHA256],
        ciphers=[pgpy.constants.SymmetricKeyAlgorithm.AES256],
        compression=[pgpy.constants.CompressionAlgorithm.ZLIB],
    )
    pub_bytes = str(key.pubkey).encode()

    mock_wkd = mock.Mock()
    mock_wkd.fetch.return_value = pub_bytes
    ctrl = ApprisePGPController(path=str(tmpdir), wkd=mock_wkd)

    # Seed the parsed-key cache with an already-expired entry
    cache_key = hashlib.sha1(b"wkd:user@example.com").hexdigest()
    # Use the private name-mangled attribute
    ctrl._ApprisePGPController__key_lookup[cache_key] = {
        "public_key": None,
        "expires": datetime.now(timezone.utc) - timedelta(seconds=1),
    }

    # The expired entry is discarded; WKD is fetched and result returned
    result = ctrl._fetch_wkd_key("user@example.com")
    assert result is not None
    mock_wkd.fetch.assert_called_once_with("user@example.com")


@pytest.mark.skipif("pgpy" not in sys.modules, reason="Requires PGPy")
def test_fetch_wkd_key_prefers_recipient_over_sender(tmpdir):
    """_fetch_wkd_key() tries recipient emails before self.email so that
    the recipient's public key (not the sender's) is used for encryption."""
    import pgpy

    # Generate a key to represent the recipient's WKD result
    key = pgpy.PGPKey.new(
        pgpy.constants.PubKeyAlgorithm.RSAEncryptOrSign, 2048
    )
    uid = pgpy.PGPUID.new("Recipient", email="recipient@example.com")
    key.add_uid(
        uid,
        usage={
            pgpy.constants.KeyFlags.Sign,
            pgpy.constants.KeyFlags.EncryptCommunications,
        },
        hashes=[pgpy.constants.HashAlgorithm.SHA256],
        ciphers=[pgpy.constants.SymmetricKeyAlgorithm.AES256],
        compression=[pgpy.constants.CompressionAlgorithm.ZLIB],
    )
    pub_bytes = str(key.pubkey).encode()

    # WKD returns a key only for the recipient address
    def fake_fetch(email):
        if email == "recipient@example.com":
            return pub_bytes
        return None

    mock_wkd = mock.Mock(side_effect=fake_fetch)
    mock_wkd.fetch.side_effect = fake_fetch
    ctrl = ApprisePGPController(
        path=str(tmpdir),
        email="sender@example.com",
        wkd=mock_wkd,
    )

    # The recipient's email is passed as a positional argument
    result = ctrl._fetch_wkd_key("recipient@example.com")
    assert result is not None
    # sender@example.com must NOT have been tried before the recipient hit
    calls = [call.args[0] for call in mock_wkd.fetch.call_args_list]
    assert calls[0] == "recipient@example.com"
    assert "sender@example.com" not in calls


@pytest.mark.skipif(
    pgp_module.PGP_SUPPORT, reason="Tests behavior when pgpy is absent"
)
def test_fetch_wkd_key_pgpy_not_installed(tmpdir):
    """_fetch_wkd_key() skips silently when pgpy raises NameError."""
    mock_wkd = mock.Mock()
    mock_wkd.fetch.return_value = b"fake-key-bytes"
    ctrl = ApprisePGPController(path=str(tmpdir), wkd=mock_wkd)
    # In a minimal env pgpy is not installed, so pgpy.PGPKey.from_blob
    # raises NameError naturally -- no mock needed.
    result = ctrl._fetch_wkd_key("user@example.com")
    assert result is None


@pytest.mark.skipif(not pgp_module.PGP_SUPPORT, reason="Requires PGPy")
def test_fetch_wkd_key_corrupt_data(tmpdir):
    """_fetch_wkd_key() skips and tries next candidate on parse failure."""
    mock_wkd = mock.Mock()
    mock_wkd.fetch.return_value = b"this-is-not-a-pgp-key"
    ctrl = ApprisePGPController(path=str(tmpdir), wkd=mock_wkd)

    # Patch the canonical pgpy module directly (not the pgp module's copy)
    # so the mock works whether pgpy was imported by name or attribute.
    with mock.patch(
        "pgpy.PGPKey.from_blob",
        side_effect=Exception("bad key"),
    ):
        result = ctrl._fetch_wkd_key("user@example.com")

    assert result is None


def test_fetch_wkd_key_uses_self_email_as_candidate(tmpdir):
    """_fetch_wkd_key() includes self.email in the candidate list."""
    mock_wkd = mock.Mock()
    mock_wkd.fetch.return_value = None
    ctrl = ApprisePGPController(
        path=str(tmpdir), email="sender@example.com", wkd=mock_wkd
    )

    ctrl._fetch_wkd_key()

    # self.email must have been tried
    mock_wkd.fetch.assert_called_with("sender@example.com")


@pytest.mark.skipif("pgpy" not in sys.modules, reason="Requires PGPy")
def test_public_key_falls_through_to_wkd(tmpdir):
    """public_key() uses WKD when no local key file exists."""
    import pgpy

    # Generate a minimal key for the mock WKD response
    key = pgpy.PGPKey.new(
        pgpy.constants.PubKeyAlgorithm.RSAEncryptOrSign, 2048
    )
    uid = pgpy.PGPUID.new("WKD User", email="wkd@example.com")
    key.add_uid(
        uid,
        usage={
            pgpy.constants.KeyFlags.Sign,
            pgpy.constants.KeyFlags.EncryptCommunications,
        },
        hashes=[pgpy.constants.HashAlgorithm.SHA256],
        ciphers=[pgpy.constants.SymmetricKeyAlgorithm.AES256],
        compression=[pgpy.constants.CompressionAlgorithm.ZLIB],
    )
    pub_bytes = str(key.pubkey).encode()

    mock_wkd = mock.Mock()
    mock_wkd.fetch.return_value = pub_bytes
    ctrl = ApprisePGPController(path=str(tmpdir), wkd=mock_wkd)

    # No local key file exists, so WKD must be tried
    result = ctrl.public_key("wkd@example.com", autogen=False)
    assert result is not None
    mock_wkd.fetch.assert_called()


@pytest.mark.skipif(not pgp_module.PGP_SUPPORT, reason="Requires PGPy")
def test_public_key_wkd_fails_falls_through_to_autogen(tmpdir):
    """public_key() falls through to autogen when WKD also returns None."""
    mock_wkd = mock.Mock()
    mock_wkd.fetch.return_value = None
    ctrl = ApprisePGPController(path=str(tmpdir), wkd=mock_wkd)
    # No local key file exists; WKD fails; autogen=False => None returned
    result = ctrl.public_key("nobody@example.com", autogen=False)
    assert result is None
    mock_wkd.fetch.assert_called()


@pytest.mark.skipif(not pgp_module.PGP_SUPPORT, reason="Requires PGPy")
def test_public_key_malformed_local_file_returns_none(tmpdir):
    """public_key() returns None gracefully when the local .asc file is
    not valid PGP data (pgpy raises during parse)."""
    # Write a file with non-PGP content under a name that public_keyfile()
    # will discover automatically
    bad_key_path = tmpdir.join("pub.asc")
    bad_key_path.write("this is not a pgp key\n")

    ctrl = ApprisePGPController(path=str(tmpdir))
    # No autogen so the only candidate is the malformed file above
    result = ctrl.public_key(autogen=False)
    assert result is None


# ---------------------------------------------------------------------------
# prune() -- WKD delegation
# ---------------------------------------------------------------------------


def test_prune_delegates_to_wkd(tmpdir):
    """prune() calls prune() on the WKD controller when one is set."""
    mock_wkd = mock.Mock()
    ctrl = ApprisePGPController(path=str(tmpdir), wkd=mock_wkd)
    ctrl.prune()
    mock_wkd.prune.assert_called_once()


def test_prune_without_wkd(tmpdir):
    """prune() runs without error when no WKD controller is set."""
    ctrl = ApprisePGPController(path=str(tmpdir))
    ctrl.prune()  # must not raise


# ---------------------------------------------------------------------------
# PGP_SUPPORT flag consistency
# ---------------------------------------------------------------------------


def test_pgp_support_flag_present():
    """PGP_SUPPORT is a boolean exported from the pgp module."""
    assert isinstance(pgp_module.PGP_SUPPORT, bool)


# ---------------------------------------------------------------------------
# private_keyfile() -- path resolution
# ---------------------------------------------------------------------------


def test_private_keyfile_no_path():
    """private_keyfile() returns None when neither path nor prv_keyfile set."""
    ctrl = ApprisePGPController(path=None)
    assert ctrl.private_keyfile() is None


def test_private_keyfile_explicit_valid(tmpdir):
    """private_keyfile() returns a path for a valid explicit prv_keyfile."""
    ctrl = ApprisePGPController(path=str(tmpdir), prv_keyfile=_VALID_PRV_ASC)
    result = ctrl.private_keyfile()
    # Should be an existing file path
    assert result and os.path.isfile(result)


def test_private_keyfile_explicit_invalid(tmpdir):
    """private_keyfile() returns False when an explicit file is unreachable."""
    ctrl = ApprisePGPController(
        path=str(tmpdir),
        prv_keyfile="/nonexistent/path/key.asc",
    )
    # The attachment itself is invalid, so the property returns False
    result = ctrl.private_keyfile()
    assert result is False


def test_private_keyfile_auto_discover_by_email(tmpdir):
    """private_keyfile() finds an auto-generated key named by email prefix."""
    # Place a fake private key named after the email prefix
    key_path = str(tmpdir.join("user-prv.asc"))
    with open(key_path, "w") as f:
        f.write("dummy")

    ctrl = ApprisePGPController(path=str(tmpdir), email="user@example.com")
    result = ctrl.private_keyfile()
    assert result == key_path


def test_private_keyfile_fallback_generic_name(tmpdir):
    """private_keyfile() falls back to pgp-prv.asc when no email-named key."""
    key_path = str(tmpdir.join("pgp-prv.asc"))
    with open(key_path, "w") as f:
        f.write("dummy")

    ctrl = ApprisePGPController(path=str(tmpdir))
    result = ctrl.private_keyfile()
    assert result == key_path


def test_private_keyfile_not_found(tmpdir):
    """private_keyfile() returns None when no private key exists on disk."""
    ctrl = ApprisePGPController(path=str(tmpdir))
    assert ctrl.private_keyfile() is None


def test_prv_keyfile_property_none_when_not_set(tmpdir):
    """prv_keyfile property returns None when no explicit key was provided."""
    ctrl = ApprisePGPController(path=str(tmpdir))
    assert ctrl.prv_keyfile is None


def test_prv_keyfile_property_path_when_valid(tmpdir):
    """prv_keyfile property returns the resolved path for a valid key."""
    ctrl = ApprisePGPController(path=str(tmpdir), prv_keyfile=_VALID_PRV_ASC)
    # Property accesses the attachment; it may return None before loading
    # because the attachment is resolved lazily -- just confirm it is not False
    assert ctrl.prv_keyfile is not False


def test_prv_keyfile_property_false_when_invalid(tmpdir):
    """prv_keyfile property returns False for an unreachable explicit file."""
    ctrl = ApprisePGPController(
        path=str(tmpdir),
        prv_keyfile="/no/such/file.asc",
    )
    assert ctrl.prv_keyfile is False


# ---------------------------------------------------------------------------
# private_key() -- loading the private PGP key object
# ---------------------------------------------------------------------------


@pytest.mark.skipif("pgpy" not in sys.modules, reason="Requires PGPy")
def test_private_key_loads_valid_fixture(tmpdir):
    """private_key() returns a PGPKey object from a valid private key file."""
    import pgpy

    ctrl = ApprisePGPController(path=str(tmpdir), prv_keyfile=_VALID_PRV_ASC)
    key = ctrl.private_key()
    assert key is not None
    assert isinstance(key, pgpy.PGPKey)


@pytest.mark.skipif("pgpy" not in sys.modules, reason="Requires PGPy")
def test_private_key_cached_on_second_call(tmpdir):
    """private_key() returns the same object on repeated calls (cached)."""
    ctrl = ApprisePGPController(path=str(tmpdir), prv_keyfile=_VALID_PRV_ASC)
    first = ctrl.private_key()
    second = ctrl.private_key()
    assert first is not None
    assert first is second


def test_private_key_returns_none_when_not_found(tmpdir):
    """private_key() returns None when no private key file exists."""
    ctrl = ApprisePGPController(path=str(tmpdir))
    assert ctrl.private_key() is None


@pytest.mark.skipif("pgpy" not in sys.modules, reason="Requires PGPy")
def test_private_key_expired_cache_reloads(tmpdir):
    """private_key() drops an expired cache entry and reloads from disk."""
    from datetime import datetime, timedelta, timezone
    import hashlib

    ctrl = ApprisePGPController(path=str(tmpdir), prv_keyfile=_VALID_PRV_ASC)

    # Load the key once to prime the cache
    first = ctrl.private_key()
    assert first is not None

    # Artificially expire the cached entry by back-dating its expiry time
    cache_key = hashlib.sha1(
        ("prv:" + os.path.abspath(_VALID_PRV_ASC)).encode("utf-8")
    ).hexdigest()
    ctrl._ApprisePGPController__key_lookup[cache_key]["expires"] = (
        datetime.now(timezone.utc) - timedelta(seconds=1)
    )

    # Second call must detect the expiry and re-load from disk
    second = ctrl.private_key()
    assert second is not None


def test_private_key_returns_none_when_pgpy_missing(tmpdir):
    """private_key() returns None gracefully when PGPy is not installed."""
    import contextlib

    ctrl = ApprisePGPController(path=str(tmpdir), prv_keyfile=_VALID_PRV_ASC)
    with contextlib.ExitStack() as stack:
        # Force PGP_SUPPORT=True so the key-loading branch is reached
        stack.enter_context(mock.patch.object(pgp_module, "PGP_SUPPORT", True))
        # Supply any readable content for the open() call
        stack.enter_context(
            mock.patch("builtins.open", mock.mock_open(read_data="data"))
        )
        if "pgpy" in sys.modules:
            # When pgpy IS installed, simulate its absence by making
            # from_blob raise NameError (same as if the name were undefined)
            stack.enter_context(
                mock.patch("pgpy.PGPKey.from_blob", side_effect=NameError)
            )
        # Without pgpy installed the NameError arises naturally at the
        # 'pgpy.PGPKey.from_blob(...)' call in private_key(); either way
        # private_key() must swallow the error and return None
        result = ctrl.private_key()
    assert result is None


def test_private_key_returns_none_when_keyfile_inaccessible(tmpdir):
    """private_key() returns None when private_keyfile() returns False."""
    # Provide an invalid path so the attachment reports the file as unreachable
    ctrl = ApprisePGPController(
        path=str(tmpdir), prv_keyfile="/no/such/file.asc"
    )
    # private_keyfile() will return False; private_key() must not raise
    result = ctrl.private_key()
    assert result is None


def test_private_key_returns_none_on_io_error(tmpdir):
    """private_key() returns None when the file cannot be read."""
    ctrl = ApprisePGPController(path=str(tmpdir), prv_keyfile=_VALID_PRV_ASC)
    with mock.patch("builtins.open", side_effect=OSError("disk error")):
        result = ctrl.private_key()
    assert result is None


def test_private_key_returns_none_on_malformed_key(tmpdir):
    """private_key() returns None when the file contains non-PGP data."""
    bad_key = str(tmpdir.join("bad-prv.asc"))
    with open(bad_key, "w") as f:
        f.write("this is not a pgp key")

    ctrl = ApprisePGPController(path=str(tmpdir), prv_keyfile=bad_key)
    result = ctrl.private_key()
    assert result is None


def test_private_key_returns_none_when_file_disappears(tmpdir):
    """private_key() handles a key file that vanishes between stat and open."""
    ctrl = ApprisePGPController(path=str(tmpdir), prv_keyfile=_VALID_PRV_ASC)
    with mock.patch("builtins.open", side_effect=FileNotFoundError("gone")):
        result = ctrl.private_key()
    assert result is None


@pytest.mark.skipif("pgpy" not in sys.modules, reason="Requires PGPy")
def test_private_key_rejects_passphrase_protected(tmpdir):
    """private_key() returns None for a passphrase-protected key."""
    import pgpy

    # Generate a fresh key and protect it with a passphrase
    key = pgpy.PGPKey.new(
        pgpy.constants.PubKeyAlgorithm.RSAEncryptOrSign, 2048
    )
    uid = pgpy.PGPUID.new("Locked", email="locked@example.com")
    key.add_uid(
        uid,
        usage={pgpy.constants.KeyFlags.Sign},
        hashes=[pgpy.constants.HashAlgorithm.SHA256],
        ciphers=[pgpy.constants.SymmetricKeyAlgorithm.AES256],
        compression=[pgpy.constants.CompressionAlgorithm.ZLIB],
    )
    key.protect(
        "s3cr3t",
        pgpy.constants.SymmetricKeyAlgorithm.AES256,
        pgpy.constants.HashAlgorithm.SHA256,
    )

    locked_path = str(tmpdir.join("locked-prv.asc"))
    with open(locked_path, "w") as f:
        f.write(str(key))

    ctrl = ApprisePGPController(path=str(tmpdir), prv_keyfile=locked_path)
    result = ctrl.private_key()
    assert result is None


# ---------------------------------------------------------------------------
# sign() -- creating detached PGP signatures
# ---------------------------------------------------------------------------


@pytest.mark.skipif("pgpy" not in sys.modules, reason="Requires PGPy")
def test_sign_returns_signature_and_micalg(tmpdir):
    """sign() returns a (sig_str, micalg) tuple for valid message + key."""
    ctrl = ApprisePGPController(path=str(tmpdir), prv_keyfile=_VALID_PRV_ASC)
    result = ctrl.sign("Hello, World!")
    assert result is not None
    sig_str, micalg = result
    # The signature must be a non-empty armored block
    assert "BEGIN PGP SIGNATURE" in sig_str
    # The micalg must start with 'pgp-'
    assert micalg.startswith("pgp-")


def test_sign_returns_none_when_no_private_key(tmpdir):
    """sign() returns None when no private key is available."""
    ctrl = ApprisePGPController(path=str(tmpdir))
    assert ctrl.sign("test message") is None


@pytest.mark.skipif("pgpy" not in sys.modules, reason="Requires PGPy")
def test_sign_returns_none_on_pgp_error(tmpdir):
    """sign() returns None when pgpy raises PGPError during signing."""
    import pgpy

    ctrl = ApprisePGPController(path=str(tmpdir), prv_keyfile=_VALID_PRV_ASC)
    with mock.patch.object(
        ctrl.private_key().__class__,
        "sign",
        side_effect=pgpy.errors.PGPError("fail"),
    ):
        result = ctrl.sign("test")
    assert result is None


def test_sign_returns_none_when_pgpy_missing(tmpdir):
    """sign() returns None gracefully when PGPy is not installed."""
    ctrl = ApprisePGPController(path=str(tmpdir), prv_keyfile=_VALID_PRV_ASC)
    # Patch private_key() to return a mock that raises NameError on sign()
    mock_key = mock.Mock()
    mock_key.sign.side_effect = NameError("pgpy not installed")
    with mock.patch.object(ctrl, "private_key", return_value=mock_key):
        result = ctrl.sign("test")
    assert result is None


# ---------------------------------------------------------------------------
# _pub_key_candidates() -- candidate filename ordering
# ---------------------------------------------------------------------------


def test_pub_key_candidates_no_email(tmpdir):
    """_pub_key_candidates() returns the four generic fallbacks when no email
    is known and no extra addresses are supplied."""
    ctrl = ApprisePGPController(path=str(tmpdir))
    # No self.email and no caller-supplied email
    candidates = ctrl._pub_key_candidates()
    # Only the four generic names should appear
    assert candidates == [
        "pgp-public.asc",
        "pgp-pub.asc",
        "public.asc",
        "pub.asc",
    ]


def test_pub_key_candidates_with_email(tmpdir):
    """_pub_key_candidates() prepends email-specific names ahead of generics,
    highest priority first (full-email before localpart)."""
    ctrl = ApprisePGPController(path=str(tmpdir), email="Chris@Example.com")
    candidates = ctrl._pub_key_candidates()
    # Full lowercase email is highest priority (index 0)
    assert candidates[0] == "chris@example.com-pub.asc"
    # Localpart shorthand is next
    assert candidates[1] == "chris-pub.asc"
    # Generic fallbacks follow
    assert "pgp-pub.asc" in candidates


def test_pub_key_candidates_extra_email(tmpdir):
    """_pub_key_candidates() gives caller-supplied addresses higher priority
    than self.email because a per-recipient key lookup should prefer the
    recipient-specific file over the controller's own identity key."""
    ctrl = ApprisePGPController(path=str(tmpdir), email="sender@example.com")
    # Simulate a lookup for a recipient address
    candidates = ctrl._pub_key_candidates("recipient@example.com")
    # Caller-supplied (recipient) address is processed last in the loop,
    # so its entries end up at the front of the list (highest priority)
    assert candidates[0] == "recipient@example.com-pub.asc"
    assert candidates[1] == "recipient-pub.asc"
    # self.email (sender) entries appear after
    assert "sender@example.com-pub.asc" in candidates
    assert "sender-pub.asc" in candidates


# ---------------------------------------------------------------------------
# _prv_key_candidates() -- candidate filename ordering
# ---------------------------------------------------------------------------


def test_prv_key_candidates_no_email(tmpdir):
    """_prv_key_candidates() returns the four generic fallbacks when no sender
    email is configured."""
    ctrl = ApprisePGPController(path=str(tmpdir))
    candidates = ctrl._prv_key_candidates()
    assert candidates == [
        "pgp-private.asc",
        "pgp-prv.asc",
        "private.asc",
        "prv.asc",
    ]


def test_prv_key_candidates_with_email(tmpdir):
    """_prv_key_candidates() prepends email-specific names ahead of generics,
    highest priority first (full-email before localpart)."""
    ctrl = ApprisePGPController(path=str(tmpdir), email="Sender@Nuxref.com")
    candidates = ctrl._prv_key_candidates()
    # Full lowercase email is highest priority
    assert candidates[0] == "sender@nuxref.com-prv.asc"
    # Localpart shorthand is next
    assert candidates[1] == "sender-prv.asc"
    # Generic fallbacks follow
    assert "pgp-prv.asc" in candidates


# ---------------------------------------------------------------------------
# public_key() -- diagnostic warning when no key is found
# ---------------------------------------------------------------------------


def test_public_key_no_key_emits_path_debug(tmpdir):
    """public_key() emits a DEBUG message listing relative cache paths when
    no public key is found and autogen is disabled.  The message is debug-
    level so the caller's warning remains the only user-visible line."""
    ctrl = ApprisePGPController(path=str(tmpdir), email="user@example.com")
    logging.disable(logging.NOTSET)
    try:
        with mock.patch.object(pgp_module, "PGP_SUPPORT", True):
            result = ctrl.public_key("user@example.com", autogen=False)
    finally:
        logging.disable(logging.CRITICAL)

    # Key not found; result must be None
    assert result is None


def test_public_key_no_key_no_path_emits_generic_debug():
    """public_key() emits a generic 'No PGP public key found' debug message
    in memory-only mode (no path to search, so no filenames to list)."""
    ctrl = ApprisePGPController(path=None)
    logging.disable(logging.NOTSET)
    try:
        with mock.patch.object(pgp_module, "PGP_SUPPORT", True):
            result = ctrl.public_key("user@example.com", autogen=False)
    finally:
        logging.disable(logging.CRITICAL)
    assert result is None


# ---------------------------------------------------------------------------
# private_key() -- diagnostic debug message when no key is found
# ---------------------------------------------------------------------------


def test_private_key_not_found_emits_path_debug(tmpdir):
    """private_key() emits a DEBUG message listing searched relative paths
    when no private key file exists on disk.  The message is debug-level so
    the caller's warning remains the only user-visible line."""
    ctrl = ApprisePGPController(path=str(tmpdir), email="sender@example.com")
    logging.disable(logging.NOTSET)
    try:
        result = ctrl.private_key()
    finally:
        logging.disable(logging.CRITICAL)
    assert result is None


def test_private_key_not_found_no_path_emits_generic_debug():
    """private_key() emits a generic 'No PGP private key found' debug
    message in memory-only mode (no storage path, nothing to search)."""
    ctrl = ApprisePGPController(path=None)
    logging.disable(logging.NOTSET)
    try:
        result = ctrl.private_key()
    finally:
        logging.disable(logging.CRITICAL)
    assert result is None
